import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import griddata

from foxes.utils.utm_utils import from_lonlat
from foxes.config import config, get_output_path
from foxes.output import FarmLayoutOutput
import foxes.variables as FV
import foxes.constants as FC

from .dataset_states import DatasetStates


class NEWAStates(DatasetStates):
    """
    Heterogeneous ambient states in NEWA-WRF format.

    Attributes
    ----------
    states_coord: str
        The states coordinate name in the data
    x_coord: str
        The x coordinate name in the data
    y_coord: str
        The y coordinate name in the data
    h_coord: str
        The height coordinate name in the data
    weight_ncvar: str
        Name of the weight data variable in the nc file(s)
    interpn_pars: dict, optional
        Additional parameters for scipy.interpolate.interpn
    bounds_extra_space: float or str
        The extra space, either float in m,
        or str for units of D, e.g. '2.5D'
    height_bounds: tuple, optional
        The (h_min, h_max) height bounds in m. Defaults to H +/- 0.5*D

    Examples
    --------
    Example of one of the NetCDF input files in NEWA format:

    >>>     Dimensions:      (time: 144, south_north: 165, west_east: 234, height: 15)
    >>>     Coordinates:
    >>>      * time         (time) datetime64[ns] 1kB 2006-01-04 ... 2006-01-04T23:54:00
    >>>      * south_north  (south_north) float32 660B -1.79e+05 -1.77e+05 ... 1.49e+05
    >>>      * west_east    (west_east) float32 936B -2.48e+05 -2.46e+05 ... 2.18e+05
    >>>      * height       (height) float32 60B 25.0 50.0 75.0 90.0 ... 400.0 500.0 1e+03
    >>>        XLAT         (south_north, west_east) float32 154kB ...
    >>>        XLON         (south_north, west_east) float32 154kB ...
    >>>    Data variables: (12/24)
    >>>        WS           (time, height, south_north, west_east) float32 334MB ...
    >>>        ...

    :group: input.states

    """

    def __init__(
        self,
        input_files_nc,
        time_coord="time",
        west_east_coord="west_east",
        south_north_coord="south_north",
        height_coord="height",
        xlat_coord="XLAT",
        xlon_coord="XLON",
        output_vars=None,
        var2ncvar=None,
        load_mode="fly",
        time_format=None,  # r"%Y-%m-%dT%H:%M:%S",
        bounds_extra_space=0.0,
        height_bounds=None,
        interp_pars=None,
        wrf_point_plot=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        input_files_nc: str
            The input netcdf file(s), can contain
            wildcards, e.g. 'wrfout_2025*.nc'
        time_coord: str
            The time coordinate name in the data
        west_east_coord: str
            The west-east coordinate name in the data
        south_north_coord: str
            The south-north coordinate name in the data
        height_coord: str, optional
            The height coordinate name in the data
        xlat_coord: str
            The latitude coordinate name in the data
        xlon_coord: str
            The longitude coordinate name in the data
        output_vars: list of str, optional
            The output variables to load, if None,
            the default variables are loaded
            (FV.WS, FV.WD, FV.TI, FV.RHO)
        var2ncvar: dict, optional
            A dictionary mapping foxes variable names
            to the corresponding netcdf variable names.
        load_mode: str
            The load mode, choices: preload, lazy, fly.
            preload loads all data during initialization,
            lazy lazy-loads the data using dask, and fly
            reads only states index and weights during initialization
            and then opens the relevant files again within
            the chunk calculations.
        time_format: str
            The datetime parsing format string
        bounds_extra_space: float or str, optional
            The extra space, either float in m,
            or str for units of D, e.g. '2.5D'
        height_bounds: tuple, optional
            The (h_min, h_max) height bounds in m. Defaults to H +/- 0.5*D
        kwargs: dict, optional
            Additional parameters for the base class
        interp_pars: dict, optional
            Additional parameters for scipy.interpolate.griddata,
            e.g. {'method': 'linear', 'fill_value': None, 'rescale': True}
        wrf_point_plot: str, optional
            Path to a plot file, e.g. wrf_points.png, to visualize the
            selected WRF grid points and the layout of the farm.
        kwargs: dict, optional
            Additional parameters for the base class

        """
        if output_vars is None:
            ovars = [FV.WS, FV.WD, FV.TI, FV.RHO]
        else:
            ovars = output_vars

        if var2ncvar is None:
            var2ncvar = {
                FV.WS: "WS",
                FV.WD: "WD",
                FV.TKE: "TKE",
                FV.RHO: "RHO",
            }

        super().__init__(
            data_source=input_files_nc,
            output_vars=ovars,
            var2ncvar=var2ncvar,
            time_format=time_format,
            load_mode=load_mode,
            weight_factor=None,
            **kwargs,
        )

        self.time_coord = time_coord
        self.west_east_coord = west_east_coord
        self.south_north_coord = south_north_coord
        self.height_coord = height_coord
        self.xlat_coord = xlat_coord
        self.xlon_coord = xlon_coord
        self.bounds_extra_space = bounds_extra_space
        self.height_bounds = height_bounds
        self.wrf_point_plot = wrf_point_plot
        self.interp_pars = interp_pars if interp_pars is not None else {}
        self.variables = list(set([v if v != FV.TI else FV.TKE for v in ovars]))

        self._cmap = {
            FC.STATE: self.time_coord,
            FV.X: self.west_east_coord,
            FV.Y: self.south_north_coord,
            FV.H: self.height_coord,
        }

    def preproc_first(
        self,
        algo,
        data,
        cmap,
        vars,
        bounds_extra_space,
        height_bounds,
        verbosity=0,
    ):
        """
        Preprocesses the first file.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data: xarray.Dataset
            The dataset to preprocess
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        vars: list
            The list of variable names
        bounds_extra_space: float or str, optional
            The extra space, either float in m,
            or str for units of D, e.g. '2.5D'
        height_bounds: tuple, optional
            The (h_min, h_max) height bounds in m. Defaults to H +/-
        verbosity: int
            The verbosity level, 0 = silent

        """

        super().preproc_first(
            algo,
            data,
            cmap,
            vars,
            bounds_extra_space=None,
            height_bounds=height_bounds,
            verbosity=verbosity,
        )

        lonlat = np.stack(
            (data[self.xlon_coord].values, data[self.xlat_coord].values), axis=-1
        )
        lonlat = np.moveaxis(lonlat, 0, 1)  # (y, x, 2) to (x, y, 2)
        nx, ny = lonlat.shape[:2]
        lonlat = lonlat.reshape((nx * ny, 2))
        xy = from_lonlat(lonlat)
        xy = xy.reshape((nx, ny, 2))
        nh = len(self._heights)

        # find horizontal bounds:
        if bounds_extra_space is not None:
            assert FV.X in cmap, (
                f"States '{self.name}': x coordinate '{FV.X}' not in cmap {cmap}"
            )
            assert FV.Y in cmap, (
                f"States '{self.name}': y coordinate '{FV.Y}' not in cmap {cmap}"
            )

            # if bounds and self.x_coord is not None and self.x_coord not in self.sel:
            xy_min, xy_max = algo.farm.get_xy_bounds(
                extra_space=bounds_extra_space, algo=algo
            )
            x0, x1 = xy_min[0], xy_max[0]
            y0, y1 = xy_min[1], xy_max[1]
            if verbosity > 0:
                print(
                    f"States '{self.name}': Restricting {FV.X} to bounds {x0:.2f} - {x1:.2f}"
                )
                print(
                    f"States '{self.name}': Restricting {FV.Y} to bounds {y0:.2f} - {y1:.2f}"
                )

            inds = np.argwhere(
                (xy[..., 0] >= x0)
                & (xy[..., 0] <= x1)
                & (xy[..., 1] >= y0)
                & (xy[..., 1] <= y1)
            )
            assert len(inds) > 0, (
                f"States '{self.name}': No grid points found within bounds (x0, x1)=({x0}, {x1}), (y0, y1)=({y0}, {y1})"
            )
            i0 = inds[:, 0].min()
            i1 = inds[:, 0].max()
            j0 = inds[:, 1].min()
            j1 = inds[:, 1].max()
            while True:
                self._xy = xy[i0 : i1 + 1, j0 : j1 + 1]
                if i0 > 0 and x0 < np.min(self._xy[..., 0]):
                    i0 -= 1
                elif i1 < nx - 1 and x1 > np.max(self._xy[..., 0]):
                    i1 += 1
                elif j0 > 0 and y0 < np.min(self._xy[..., 1]):
                    j0 -= 1
                elif j1 < ny - 1 and y1 > np.max(self._xy[..., 1]):
                    j1 += 1
                else:
                    break
            nx, ny = self._xy.shape[:2]

            if self.isel is None:
                self.isel = {}
            self.isel.update(
                {
                    self.west_east_coord: slice(i0, i1 + 1),
                    self.south_north_coord: slice(j0, j1 + 1),
                }
            )
            if verbosity > 0:
                print(
                    f"States '{self.name}': Selected {FV.X} = {np.min(self._xy[..., 0]):.2f} - {np.max(self._xy[..., 0]):.2f} ({nx} points)"
                )
                print(
                    f"States '{self.name}': Selected {FV.Y} = {np.min(self._xy[..., 1]):.2f} - {np.max(self._xy[..., 1]):.2f} ({ny} points)"
                )
                print(
                    f"States '{self.name}': Selected {self._xy.shape[:2] + (nh,)} grid points"
                )
        else:
            self._xy = xy
            if verbosity > 0:
                print(
                    f"States '{self.name}': Selecting all {self._xy.shape[:2] + (nh,)} grid points"
                )

        if self.wrf_point_plot is not None:
            fpath = get_output_path(self.wrf_point_plot)
            if verbosity > 0:
                print(f"States '{self.name}': Writing WRF grid point plot to '{fpath}'")
            fig, ax = plt.subplots(figsize=(8, 8))
            ax.plot(
                self._xy[..., 0].flatten(),
                self._xy[..., 1].flatten(),
                c="blue",
                alpha=0.2,
                marker=".",
                linestyle="None",
            )
            anno = 3 if len(algo.farm.wind_farm_names) > 1 else 0
            FarmLayoutOutput(farm=algo.farm).get_figure(
                fig=fig, ax=ax, annotate=anno, fontsize=12
            )
            ax.set_xlabel(f"{FV.X} [m]")
            ax.set_ylabel(f"{FV.Y} [m]")
            ax.set_aspect("equal", adjustable="box")
            ax.autoscale_view(tight=True)
            fig.savefig(fpath, bbox_inches="tight")
            plt.close()

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        return super().load_data(
            algo,
            cmap=self._cmap,
            variables=self.variables,
            bounds_extra_space=self.bounds_extra_space,
            height_bounds=self.height_bounds,
            verbosity=verbosity,
        )

    def interpolate_data(self, idims, icrds, d, pts, tdims, vrs, times):
        """
        Interpolates data to points.

        This function should be implemented in derived classes.

        Parameters
        ----------
        idims: list of str
            The input dimensions, e.g. [state, x, y, height]
        icrds: list of numpy.ndarray
            The input coordinates, each with shape (n_i,)
            where n_i is the number of grid points in dimension i
        d: numpy.ndarray
            The data array, with shape (n1, n2, ..., nv)
            where ni represents the dimension sizes and
            nv is the number of variables
        pts: numpy.ndarray
            The points to interpolate to, with shape (n_pts, n_idims)
        tdims: tuple
            The target dimensions, e.g. (1, m, nv) or (n_states, m, nv)
        vrs: list of str
            The variable names, length nv
        times: numpy.ndarray
            The time coordinates of the states, with shape (n_states,)

        Returns
        -------
        d_interp: numpy.ndarray
            The interpolated data array with shape tdims

        """

        # prepare interpolation parameters:
        ipars = dict(
            method="linear",
            rescale=True,
            fill_value=np.nan,
        )
        ipars.update(self.interp_pars)

        # prepare grid points:
        gpts = np.zeros(d.shape[:-1] + (len(idims),), dtype=config.dtype_double)
        n_gpts = 1
        ix = None
        for i, c in enumerate(icrds):
            if idims[i] not in (FV.X, FV.Y):  # leave out X and Y, cf. self._cmap
                shp = [1] * len(icrds)
                shp[i] = c.shape[0]
                gpts[..., i] = c.reshape(shp)
                n_gpts *= c.shape[0]
            elif idims[i] == FV.X:
                assert FV.Y in idims, (
                    f"States '{self.name}': {FV.X} found in dims {idims} but not {FV.Y}"
                )
                ix = i
            else:
                assert ix == i - 1, (
                    f"States '{self.name}': Unexpected dimension order {idims}, expected {FV.X} before {FV.Y}"
                )

        # sneak in self._xy instead of west_east and south_north coords:
        if ix is not None:
            shp = [1] * (len(icrds) + 1)
            shp[ix : ix + 2] = self._xy.shape[:2]
            shp[-1] = 2
            gpts[..., ix : ix + 2] = self._xy.reshape(shp)
            n_gpts *= self._xy.shape[0] * self._xy.shape[1]

        # reshape:
        gpts = gpts.reshape((n_gpts, len(idims)))
        n_vrs = d.shape[-1]
        d = d.reshape((n_gpts, n_vrs))

        # remove NaN data points:
        sel = np.any(np.isnan(d), axis=1)
        if np.all(sel):
            i = np.where(np.isnan(d))
            p = gpts[i[0][0]]
            v = vrs[i[1][0]]
            t = times[int(p[0])]
            print(f"NaN data found in input data, e.g. for variable '{v}' at point:")
            print("   time:   ", t)
            for ic, c in enumerate(idims):
                print(f"  {c}: {p[ic]}")
            for iw, w in enumerate(vrs):
                print(f"  {w}: {d[i[0][0], iw]}")
            print("\n\n")
            raise ValueError(
                f"States '{self.name}': All grid points contain NaN values during interpolation, e.g. at time {t}."
            )
        elif np.any(sel):
            gpts = gpts[~sel, :]
            d = d[~sel, :]

        # run interpolation:
        results = griddata(gpts, d, pts, **ipars)

        # check for error:
        if np.isnan(ipars.get("fill_value", np.nan)):
            sel = np.isnan(results[..., 0])
            if np.any(sel):
                i = np.where(sel)[0]
                n_dms = len(icrds)
                p = pts[i[0], :n_dms]
                t = times[int(p[0])]
                qmin = np.min(gpts[:, :n_dms], axis=0)
                qmax = np.max(gpts[:, :n_dms], axis=0)
                isin = (p >= qmin) & (p <= qmax)
                method = "linear"
                print("\n\nInterpolation error", i, p)
                print("time:   ", t)
                print("dims:   ", idims)
                print("point:  ", p)
                print("qmin:   ", qmin)
                print("qmax:   ", qmax)
                print("Inside: ", isin, "\n\n")

                if not np.all(isin):
                    raise ValueError(
                        f"States '{self.name}': Interpolation method '{method}' failed for {np.sum(sel)} points, e.g. for point {p} at time {t}, outside of bounds {qmin} - {qmax}, dimensions = {idims}. "
                    )
                else:
                    sel2 = np.isnan(d)
                    if np.any(sel2):
                        i = np.where(sel2)
                        p = gpts[i[0][0]]
                        v = vrs[i[1][0]]
                        print(
                            f"NaN data found in input data during interpolation, e.g. for variable '{v}' at point:"
                        )
                        print("   time:   ", t)
                        for ic, c in enumerate(idims):
                            print(f"  {c}: {p[ic]}")
                        for iw, w in enumerate(vrs):
                            print(f"  {w}: {d[i[0][0], iw]}")
                        print("\n\n")
                        raise ValueError(
                            f"States '{self.name}': Interpolation method '{method}' failed, NaN values found in input data for {np.sum(sel)} grid points, e.g. {gpts[i[0]]} at time {t} with {v} = {d[i[0][0], i[1][0]]}."
                        )
                    raise ValueError(
                        f"States '{self.name}': Interpolation method '{method}' failed for {np.sum(sel)} points, for unknown reason."
                    )
        return results.reshape(tdims)

    def calculate(self, algo, mdata, fdata, tdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        results = super().calculate(algo, mdata, fdata, tdata)

        # convert TKE to TI if needed:
        if FV.TI in self.ovars and FV.TI not in results:
            assert FV.WS in results, (
                f"States '{self.name}': Cannot calculate {FV.TI} without {FV.WS}"
            )
            assert FV.TKE in results or FV.TKE in self.ovars, (
                f"States '{self.name}': Cannot calculate {FV.TI} without {FV.TKE}"
            )
            if FV.TKE not in self.ovars:
                tke = results.pop(FV.TKE)
            else:
                tke = results[FV.TKE]
            ws = results[FV.WS]
            results[FV.TI] = np.sqrt(1.5 * tke) / ws

        return results



from foxes.algorithms.sequential import SequentialPlugin

from ..flow_plots_2d.flow_plots import FlowPlots2D


class SeqFlowAnimationPlugin(SequentialPlugin):
    """
    Plugin for creating data for a 2D flow animation
    during sequential iterations

    Attributes
    ----------
    orientation: str
        The orientation, either "yx", "xz" or "yz"
    pars: dict
        Additional parameters for plotting

    :group: output.seq_plugins

    """

    def __init__(self, orientation, **pars):
        """
        Constructor.

        Parameters
        ----------
        orientation: str
            The orientation, either "yx", "xz" or "yz"
        pars: dict, optional
            Additional parameters for plotting

        """
        super().__init__()
        self.orientation = orientation
        self.pars = pars

        if "title" in self.pars and callable(self.pars["title"]):
            self._tfun = self.pars.pop("title")
        else:
            self._tfun = None

    def initialize(self, algo):
        """
        Initialize data based on the intial iterator

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The current sequential algorithm

        """
        super().initialize(algo)
        self._data = []

    def update(self, algo, fres, pres=None):
        """
        Updates data based on current iteration

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The latest sequential algorithm
        fres: xarray.Dataset
            The latest farm results
        pres: xarray.Dataset, optional
            The latest point results

        """
        super().update(algo, fres, pres)

        o = FlowPlots2D(algo, fres)

        if self._tfun is not None:
            self.pars["title"] = self._tfun(algo.states.counter, algo.states.index()[0])

        if self.orientation == "xy":
            d = next(o.gen_states_fig_xy(**self.pars, precalc=True))
        elif self.orientation == "xz":
            d = next(o.gen_states_fig_xz(**self.pars, precalc=True))
        elif self.orientation == "yz":
            d = next(o.gen_states_fig_yz(**self.pars, precalc=True))
        else:
            raise KeyError(
                f"Unkown orientation '{self.orientation}', choises: xy, xz, yz"
            )

        # minimize stored data:
        od = [d[0], d[1], None]
        if len(self._data) == 0:
            od[2] = d[2]
        of = (
            fres
            if ("rotor_color" in self.pars and self.pars["rotor_color"] is not None)
            else None
        )

        self._data.append((of, od))

    def gen_images(self, ax):
        """

        Parameters
        ----------
        ax: matplotlib.Axis
            The plotting axis

        Yields
        ------
        imgs: tuple
            The (figure, artists) tuple

        """
        fig = ax.get_figure()
        gdata = None
        while len(self._data):
            fres, d = self._data.pop(0)

            if d[2] is not None:
                gdata = d[2]

            o = FlowPlots2D(self.algo, fres)

            yield next(
                o.gen_states_fig_xy(
                    **self.pars,
                    ax=ax,
                    fig=fig,
                    ret_im=True,
                    precalc=(d[0], d[1], gdata),
                )
            )

            del o, fres, d

            if (
                self.pars.get("vmin", None) is not None
                and self.pars.get("vmax", None) is not None
            ):
                self.pars["add_bar"] = False



import numpy as np
import pandas as pd
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV


def calc(args, rotor, sdata, pwake):
    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype
    D = ttype.D

    farm = foxes.WindFarm()
    farm.add_turbine(
        foxes.Turbine(
            xy=np.array([0.0, 0.0]), turbine_models=args.tmodels + [ttype.name]
        ),
        verbosity=0,
    )
    farm.add_turbine(
        foxes.Turbine(
            xy=np.array([args.dist_x, 0.0]),
            turbine_models=["sety"] + args.tmodels + [ttype.name],
        ),
        verbosity=0,
    )

    states = foxes.input.states.StatesTable(
        sdata,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti", FV.RHO: "rho"},
    )

    ydata = np.full((len(sdata.index), farm.n_turbines), np.nan)
    ydata[:, 1] = sdata["y"].to_numpy()
    mbook.turbine_models["sety"] = foxes.models.turbine_models.SetFarmVars(
        pre_rotor=True
    )
    mbook.turbine_models["sety"].add_var(FV.Y, ydata)

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        wake_models=args.wakes,
        rotor_model=rotor,
        wake_frame="rotor_wd",
        partial_wakes=pwake,
        mbook=mbook,
        verbosity=0,
    )

    print(f"\nCalculating rotor = {rotor}, pwake = {pwake}")
    farm_results = algo.calc_farm()

    return farm_results, D


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.08)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-v", "--var", help="The variable selection", default=FV.REWS)
    parser.add_argument(
        "-d", "--dist_x", help="The turbine distance in x", type=float, default=500.0
    )
    parser.add_argument(
        "-y0", "--ymin", help="The minimal y value", type=float, default=-500.0
    )
    parser.add_argument(
        "-y1", "--ymax", help="The maximal y value", type=float, default=500.0
    )
    parser.add_argument(
        "-ys", "--ystep", help="The step size in y direction", type=float, default=1.0
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_k002"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=["kTI_02"], nargs="+"
    )
    parser.add_argument(
        "-r", "--rotors", help="The rotor model(s)", default=["grid400"], nargs="+"
    )
    parser.add_argument(
        "-p",
        "--pwakes",
        help="The partial wakes model(s)",
        default=["grid16", "axiwake6", "rotor_points"],
        nargs="+",
    )
    parser.add_argument("-tt", "--title", help="The figure title", default=None)
    parser.add_argument(
        "-e",
        "--engine",
        help="The engine",
        default="ProcessEngine",
    )
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    ws = args.ws
    var = args.var
    swks = ", ".join(args.wakes)
    ttl0 = f"ws$_0$ = {ws} m, ti$_0$ = {args.ti}"

    varn = 1
    vlab = var
    if var in [FV.WS, FV.REWS, FV.REWS2, FV.REWS3]:
        varn = ws
        vlab = f"{var}/ws$_0$"

    Ny = int((args.ymax - args.ymin) // args.ystep)
    sdata = pd.DataFrame(index=range(Ny + 1))
    sdata.index.name = "state"
    sdata["ws"] = args.ws
    sdata["wd"] = args.wd
    sdata["ti"] = args.ti
    sdata["rho"] = args.rho
    sdata["y"] = np.linspace(args.ymin, args.ymax, Ny + 1)

    fig, ax = plt.subplots(figsize=(10, 4))
    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        if len(args.rotors) == 1:
            for pwake in args.pwakes:
                farm_results, D = calc(args, args.rotors[0], sdata, pwake)

                ax.plot(
                    farm_results[FV.Y][:, 1] / D,
                    farm_results[var][:, 1] / varn,
                    linewidth=2,
                    alpha=0.6,
                    label=pwake,
                )

                title = f"{swks}, variable {var}\nVarying partial wake models, {ttl0}, rotor = {args.rotors[0]}"

        elif len(args.pwakes) == 1:
            for rotor in args.rotors:
                farm_results, D = calc(args, rotor, sdata, args.pwakes[0])

                ax.plot(
                    farm_results[FV.Y][:, 1] / D,
                    farm_results[var][:, 1] / varn,
                    linewidth=2,
                    alpha=0.6,
                    label=rotor,
                )

                title = f"{swks}, variable {var}\nVarying rotor models, {ttl0}, pwake = {args.pwakes[0]}"

        elif len(args.rotors) == len(args.pwakes):
            for rotor, pwake in zip(args.rotors, args.pwakes):
                farm_results, D = calc(args, rotor, sdata, pwake)

                ax.plot(
                    farm_results[FV.Y][:, 1] / D,
                    farm_results[var][:, 1] / varn,
                    linewidth=2,
                    alpha=0.6,
                    label=f"{rotor}, {pwake}",
                )

                title = "{swks}, variable {var}\nVarying rotor and partial wake models, {ttl0}"

        else:
            raise ValueError(
                "Please either give one rotor, or one pwake, or same number of both"
            )

    if args.title is not None:
        title = args.title

    ax.set_title(title)
    ax.set_xlabel("y/D")
    ax.set_ylabel(vlab)
    ax.legend()
    if not args.nofig:
        plt.show()



import foxes

if __name__ == "__main__":
    states = foxes.input.states.Timeseries(
        "timeseries_3000.csv.gz", ["WS", "WD", "TI", "RHO"]
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm, "test_farm_67.csv", turbine_models=["NREL5MW"]
    )

    algo = foxes.algorithms.Downwind(farm, states, ["Jensen_linear_k007"])
    farm_results = algo.calc_farm()

    print(farm_results)



import pandas as pd
from pathlib import Path
import inspect

import foxes
import foxes.variables as FV

thisdir = Path(inspect.getfile(inspect.currentframe())).parent


def test():
    print(thisdir)

    cfile = thisdir / "flappy" / "results.csv.gz"
    tfile = thisdir / "NREL-5MW-D126-H90.csv"
    sfile = thisdir / "states.csv.gz"
    lfile = thisdir / "test_farm.csv"

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(
        data_source=tfile, var_ws_ct=FV.REWS, var_ws_P=FV.REWS
    )
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.StatesTable(
        data_source=sfile,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti"},
        fixed_vars={FV.RHO: 1.225},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm, lfile, turbine_models=[ttype.name], verbosity=0
    )

    with foxes.Engine.new("threads", chunk_size_states=1000):
        algo = foxes.algorithms.Downwind(
            farm,
            states,
            mbook=mbook,
            rotor_model="centre",
            wake_models=["Jensen_linear_k007"],
            wake_frame="rotor_wd",
            partial_wakes={"Jensen_linear_k007": "top_hat"},
            verbosity=0,
        )

        data = algo.calc_farm()

        df = data.to_dataframe()[
            [FV.AMB_WD, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]
        ]

        print()
        print("TRESULTS\n")
        print(df)

        print("\nReading file", cfile)
        fdata = pd.read_csv(cfile)
        print(fdata)

        print("\nVERIFYING\n")
        df[FV.WS] = df["REWS"]
        df[FV.AMB_WS] = df["AMB_REWS"]

        delta = (df.reset_index() - fdata)[[FV.AMB_WS, FV.AMB_P, FV.WS, FV.P]]
        print("\nDELTA\n", delta.describe())

        chk = delta.abs().max()
        print("\nCHK\n", chk)
        print("CHK WS =", chk[FV.WS])
        print("CHK P =", chk[FV.P])

        assert chk[FV.WS] < 1e-5
        assert chk[FV.P] < 1e-3


if __name__ == "__main__":
    test()



import numpy as np
from abc import ABC
from itertools import count

from foxes.config import config
import foxes.constants as FC


class Model(ABC):
    """
    Base class for all models.

    Attributes
    ----------
    name: str
        The model name

    :group: core

    """

    _ids = {}

    def __init__(self):
        """
        Constructor.
        """
        t = type(self).__name__
        if t not in self._ids:
            self._ids[t] = count(0)
        self._id = next(self._ids[t])

        self.name = f"{type(self).__name__}"
        if self._id > 0:
            self.name += f"_instance{self._id}"

        self.__initialized = False
        self.__running = False

    def __repr__(self):
        return f"{type(self).__name__}()"

    @property
    def model_id(self):
        """
        Unique id based on the model type.

        Returns
        -------
        int
            Unique id of the model object

        """
        return self._id

    def var(self, v):
        """
        Creates a model specific variable name.

        Parameters
        ----------
        v: str
            The variable name

        Returns
        -------
        vnm: str
            Model specific variable name

        """
        return f"{self.name}_{v}"

    def unvar(self, vnm):
        """
        Translates model specific variable name to origninal variable name.

        Parameters
        ----------
        vnm: str
            The vamodel specific variable name

        Returns
        -------
        v: str
            Original variable name

        """
        lng = len(f"{self.name}_")
        return vnm[lng:] if vnm.startswith(f"{self.name}_") else None

    @property
    def initialized(self):
        """
        Initialization flag.

        Returns
        -------
        bool:
            True if the model has been initialized.

        """
        return self.__initialized

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return []

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        if self.initialized:
            raise ValueError(
                f"Model '{self.name}': Cannot call load_data after initialization"
            )
        return {"coords": {}, "data_vars": {}}

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if self.running:
            raise ValueError(f"Model '{self.name}': Cannot initialize while running")
        if not self.initialized:
            pr = False
            for m in self.sub_models():
                if not m.initialized:
                    if verbosity > 1 and not pr:
                        print(f">> {self.name}: Starting sub-model initialization >> ")
                        pr = True
                    m.initialize(algo, verbosity)
            if pr:
                print(f"<< {self.name}: Finished sub-model initialization << ")

            if verbosity > 0:
                print(f"Initializing model '{self.name}'")
            algo.store_model_data(self, self.load_data(algo, verbosity), force)

            self.__initialized = True

    @property
    def running(self):
        """
        Flag for currently running models

        Returns
        -------
        flag: bool
            True if currently running

        """
        return self.__running

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        if self.running:
            raise ValueError(
                f"Model '{self.name}': Cannot call set_running while running"
            )
        for m in self.sub_models():
            if not m.running:
                m.set_running(algo, data_stash, sel, isel, verbosity=verbosity)

        if verbosity > 0:
            print(f"Model '{self.name}': running")
        if data_stash is not None and self.name not in data_stash:
            data_stash[self.name] = {}

        self.__running = True

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        if not self.running:
            raise ValueError(
                f"Model '{self.name}': Cannot call unset_running when not running"
            )
        for m in self.sub_models():
            if m.running:
                m.unset_running(algo, data_stash, sel, isel, verbosity=verbosity)

        if verbosity > 0:
            print(f"Model '{self.name}': not running")
        self.__running = False

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if self.running:
            raise ValueError(f"Model '{self.name}': Cannot finalize while running")
        if self.initialized:
            pr = False
            for m in self.sub_models():
                if verbosity > 1 and not pr:
                    print(f">> {self.name}: Starting sub-model finalization >> ")
                    pr = True
                m.finalize(algo, verbosity)
            if pr:
                print(f"<< {self.name}: Finished sub-model finalization << ")

            if verbosity > 0:
                print(f"Finalizing model '{self.name}'")
            algo.del_model_data(self)

            self.__initialized = False

    def get_data(
        self,
        variable,
        target,
        lookup="smfp",
        mdata=None,
        fdata=None,
        tdata=None,
        downwind_index=None,
        accept_none=False,
        accept_nan=True,
        algo=None,
        upcast=False,
        selection=None,
    ):
        """
        Getter for a data entry in the model object
        or provided data sources

        Parameters
        ----------
        variable: str
            The variable, serves as data key
        target: str, optional
            The dimensions identifier for the output,
            FC.STATE_TURBINE, FC.STATE_TARGET or
            FC.STATE_TARGET_TPOINT
        lookup: str
            The order of data sources. Combination of:
            's' for self,
            'm' for mdata,
            'f' for fdata,
            't' for tdata,
            'w' for wake modelling data
        mdata: foxes.core.Data, optional
            The model data
        fdata: foxes.core.Data, optional
            The farm data
        tdata: foxes.core.Data, optional
            The target point data
        downwind_index: int, optional
            The index in the downwind order
        data_prio: bool
            First search the data source, then the object
        accept_none: bool
            Do not throw an error if data entry is None
        accept_nan: bool
            Do not throw an error if data entry is np.nan
        algo: foxes.core.Algorithm, optional
            The algorithm, needed for data from previous iteration
        upcast: bool
            Flag for ensuring targets dimension,
            otherwise dimension 1 is entered
        selection: numpy.ndarray, optional
            Apply this selection to the result,
            state-turbine, state-target, or state-target-tpoint

        """

        def _geta(a):
            sources = [s for s in [mdata, fdata, tdata, algo, self] if s is not None]
            for s in sources:
                try:
                    if a == "states_i0":
                        out = s.states_i0(counter=True)
                        if out is not None:
                            return out
                    else:
                        out = getattr(s, a)
                        if out is not None:
                            return out
                except AttributeError:
                    pass
            raise KeyError(
                f"Model '{self.name}': Failed to determine '{a}'. Maybe add to arguments of get_data: mdata, fdata, tdata, algo?"
            )

        n_states = _geta("n_states")
        if target == FC.STATE_TURBINE:
            if downwind_index is not None:
                raise ValueError(
                    f"Target '{target}' is incompatible with downwind_index (here {downwind_index})"
                )
            n_turbines = _geta("n_turbines")
            dims = (FC.STATE, FC.TURBINE)
            shp = (n_states, n_turbines)
        elif target == FC.STATE_TARGET:
            n_targets = _geta("n_targets")
            dims = (FC.STATE, FC.TARGET)
            shp = (n_states, n_targets)
        elif target == FC.STATE_TARGET_TPOINT:
            n_targets = _geta("n_targets")
            n_tpoints = _geta("n_tpoints")
            dims = (FC.STATE, FC.TARGET, FC.TPOINT)
            shp = (n_states, n_targets, n_tpoints)
        else:
            raise KeyError(
                f"Model '{self.name}': Wrong parameter 'target = {target}'. Choices: {FC.STATE_TURBINE}, {FC.STATE_TARGET}, {FC.STATE_TARGET_TPOINT}"
            )

        def _match_shape(a):
            out = np.asarray(a)
            if len(out.shape) < len(shp):
                for i, s in enumerate(shp):
                    if i >= len(out.shape):
                        out = out[..., None]
                    elif a.shape[i] not in (1, s):
                        raise ValueError(
                            f"Shape mismatch for '{variable}': Got {out.shape}, expecting {shp}"
                        )
            elif len(out.shape) > len(shp):
                raise ValueError(
                    f"Shape mismatch for '{variable}': Got {out.shape}, expecting {shp}"
                )
            return out

        def _filter_dims(source):
            a = source[variable]
            a_dims = tuple(source.dims[variable])
            if downwind_index is None or FC.TURBINE not in a_dims:
                d = a_dims
            else:
                slc = tuple(
                    [downwind_index if dd == FC.TURBINE else np.s_[:] for dd in a_dims]
                )
                a = a[slc]
                d = tuple([dd for dd in a_dims if dd != FC.TURBINE])
            return a, d

        out = None
        for s in lookup:
            # lookup self:
            if s == "s" and hasattr(self, variable):
                a = getattr(self, variable)
                if a is not None:
                    out = _match_shape(a)

            # lookup mdata:
            elif s == "m" and mdata is not None and variable in mdata:
                a, d = _filter_dims(mdata)
                ld = len(d)
                if ld <= len(dims) and d == dims[:ld]:
                    out = _match_shape(mdata[variable])

            # lookup fdata:
            elif (
                s == "f"
                and fdata is not None
                and variable in fdata
                and tuple(fdata.dims[variable]) == (FC.STATE, FC.TURBINE)
            ):
                if target == FC.STATE_TURBINE:
                    out = fdata[variable]
                elif downwind_index is not None:
                    out = _match_shape(fdata[variable][:, downwind_index])

            # lookup pdata:
            elif (
                s == "t"
                and target != FC.STATE_TURBINE
                and tdata is not None
                and variable in tdata
            ):
                a, d = _filter_dims(tdata)
                ld = len(ld)
                if ld <= len(dims) and d == dims[:ld]:
                    out = _match_shape(tdata[variable])

            # lookup wake modelling data:
            elif (
                s == "w"
                and fdata is not None
                and tdata is not None
                and variable in fdata
                and tuple(fdata.dims[variable]) == (FC.STATE, FC.TURBINE)
                and downwind_index is not None
                and algo is not None
            ):
                out = _match_shape(
                    algo.wake_frame.get_wake_modelling_data(
                        algo,
                        variable,
                        downwind_index,
                        fdata,
                        tdata=tdata,
                        target=target,
                    )
                )

            if out is not None:
                break

        # check for None:
        if out is None:
            if not accept_none:
                raise ValueError(
                    f"Model '{self.name}': Variable '{variable}' is requested but not found."
                )
            return out

        # data from other chunks, only with iterations:
        if (
            target in [FC.STATE_TARGET, FC.STATE_TARGET_TPOINT]
            and fdata is not None
            and variable in fdata
            and tdata is not None
            and FC.STATES_SEL in tdata
        ):
            if out.shape != shp:
                # upcast to dims:
                tmp = np.zeros(shp, dtype=out.dtype)
                tmp[:] = out
                out = tmp
                del tmp
            else:
                out = out.copy()
            if downwind_index is None:
                raise KeyError(
                    f"Model '{self.name}': Require downwind_index for obtaining results from previous iteration"
                )
            if tdata[FC.STATE_SOURCE_ORDERI] != downwind_index:
                raise ValueError(
                    f"Model '{self.name}': Expecting downwind_index {tdata[FC.STATE_SOURCE_ORDERI]}, got {downwind_index}"
                )
            if algo is None:
                raise ValueError(
                    f"Model '{self.name}': Iteration data found for variable '{variable}', requiring algo"
                )

            from foxes.algorithms.sequential import Sequential

            if isinstance(algo, Sequential):
                i0 = algo.states.counter
            else:
                i0 = _geta("states_i0")
            sts = tdata[FC.STATES_SEL]
            if target == FC.STATE_TARGET and tdata.n_tpoints != 1:
                # find the mean index and round it to nearest integer:
                sts = tdata.tpoint_mean(FC.STATES_SEL)[:, :, None]
                sts = (sts + 0.5).astype(config.dtype_int)
            sel = sts < i0
            if np.any(sel):
                if not hasattr(algo, "farm_results_downwind"):
                    raise KeyError(
                        f"Model '{self.name}': Iteration data found for variable '{variable}', requiring iterative algorithm"
                    )
                prev_fres = getattr(algo, "farm_results_downwind")
                if prev_fres is not None:
                    prev_data = prev_fres[variable].to_numpy()[sts[sel], downwind_index]
                    if target == FC.STATE_TARGET:
                        out[sel[:, :, 0]] = prev_data
                    else:
                        out[sel] = prev_data
                    del prev_data
                del prev_fres
            if np.any(~sel):
                sts = sts[~sel] - i0
                sel_data = fdata[variable][sts, downwind_index]
                if target == FC.STATE_TARGET:
                    out[~sel[:, :, 0]] = sel_data
                else:
                    out[~sel] = sel_data
                del sel_data
            del sel, sts

        # check for nan:
        if not accept_nan:
            try:
                if np.all(np.isnan(np.atleast_1d(out))):
                    raise ValueError(
                        f"Model '{self.name}': Requested variable '{variable}' contains NaN values."
                    )
            except TypeError:
                pass

        # apply selection:
        if selection is not None:

            def _upcast_sel(sel_shape):
                chp = []
                for i, s in enumerate(out.shape):
                    if i < len(sel_shape) and sel_shape[i] > 1:
                        if sel_shape[i] != shp[i]:
                            raise ValueError(
                                f"Incompatible selection shape {sel_shape} for output shape {shp[i]}"
                            )
                        chp.append(shp[i])
                    else:
                        chp.append(s)
                chp = tuple(chp)
                eshp = list(shp[len(sel_shape) :])
                if chp != out.shape:
                    nout = np.zeros(chp, dtype=out.dtype)
                    nout[:] = out
                    return nout, eshp
                return out, eshp

            if isinstance(selection, np.ndarray) and selection.dtype == bool:
                if len(selection.shape) > len(out.shape):
                    raise ValueError(
                        f"Expecting selection of shape {out.shape}, got {selection.shape}"
                    )
                out, eshp = _upcast_sel(selection.shape)
            elif isinstance(selection, (tuple, list)):
                if len(selection) > len(out.shape):
                    raise ValueError(
                        f"Selection is tuple/list of length {len(selection)}, expecting <= {len(out.shape)} "
                    )
                out, eshp = _upcast_sel(shp[: len(selection)])
            else:
                raise TypeError(
                    f"Expecting selection of type np.ndarray (bool), or tuple, or list. Got {type(selection).__name__}"
                )
            out = out[selection]
            shp = tuple([len(out)] + list(eshp))

        # apply upcast:
        if upcast and out.shape != shp:
            tmp = np.zeros(shp, dtype=out.dtype)
            tmp[:] = out
            out = tmp
            del tmp

        return out



from copy import deepcopy

from foxes.utils import import_module

from .pool import PoolEngine


ray = None


def load_ray():
    """On-demand loading of the ray package"""
    global ray
    if ray is None:
        ray = import_module("ray")


class RayEngine(PoolEngine):
    """
    The ray engine for foxes calculations.

    :group: engines

    """

    def _create_pool(self):
        """Creates the pool"""
        self.print(f"Initializing pool of {self.n_workers} ray workers")
        load_ray()
        ray.init(num_cpus=self.n_workers)

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """

        @ray.remote
        def f_ray(*args, **kwargs):
            return f(*deepcopy(args), **deepcopy(kwargs))

        return f_ray.remote(*args, **kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return ray.get(future)

    def _shutdown_pool(self):
        """Shuts down the pool"""
        self.print(f"Shutting down pool of {self.n_workers} ray workers")
        ray.shutdown()



from foxes.utils import import_module

from .futures import ProcessEngine


class MPIEngine(ProcessEngine):
    """
    The MPI engine for foxes calculations.

    Examples
    --------
    Run command, e.g. for 12 processors and a script run.py:

    >>> mpiexec -n 12 -m mpi4py.futures run.py

    :group: engines

    """

    def _create_pool(self):
        """Creates the pool"""
        MPIPoolExecutor = import_module(
            "mpi4py.futures",
            pip_hint="pip install mpi4py",
            conda_hint="conda install mpi4py -c conda-forge",
        ).MPIPoolExecutor
        self._pool = MPIPoolExecutor(max_workers=self.n_workers)



from foxes.core import Turbine


def add_from_arrays(
    farm,
    x,
    y,
    heights=None,
    diameters=None,
    ids=None,
    names=None,
    turbine_base_name="T",
    turbine_base_name_count_shift=False,
    verbosity=1,
    **turbine_parameters,
):
    """
    Add turbines to wind farm from direct one dimensional data arrays.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    x: list or numpy.ndarray
        The x-coordinates of the turbines
    y: list or numpy.ndarray
        The y-coordinates of the turbines
    heights: list or numpy.ndarray, optional
        The hub heights of the turbines, or None
    diameters: list or numpy.ndarray, optional
        The rotor diameters of the turbines, or None
    ids: list or numpy.ndarray, optional
        The ids of the turbines, or None
    names: list or numpy.ndarray, optional
        The names of the turbines, or None
    turbine_base_name: str, optional
        The turbine base name, only used
        if col_name is None
    turbine_base_name_count_shift: bool, optional
        Start turbine names by 1 instead of 0
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Additional parameters are forwarded to the WindFarm.add_turbine().

    :group: input.farm_layout

    """
    tmodels = turbine_parameters.pop("turbine_models", [])
    H = turbine_parameters.pop("H", None)
    D = turbine_parameters.pop("D", None)

    for i in range(len(x)):
        s = 1 if turbine_base_name_count_shift else 0
        tname = f"{turbine_base_name}{i + s}" if names is None else names[i]

        farm.add_turbine(
            Turbine(
                name=tname,
                index=ids[i] if ids is not None else i,
                xy=[x[i], y[i]],
                H=heights[i] if heights is not None else H,
                D=diameters[i] if diameters is not None else D,
                turbine_models=tmodels,
                **turbine_parameters,
            ),
            verbosity=verbosity,
        )



from .set_amb_farm_results import SetAmbFarmResults as SetAmbFarmResults
from .set_amb_point_results import SetAmbPointResults as SetAmbPointResults
from .farm_wakes_calc import FarmWakesCalculation as FarmWakesCalculation
from .point_wakes_calc import PointWakesCalculation as PointWakesCalculation
from .init_farm_data import InitFarmData as InitFarmData
from .reorder_farm_output import ReorderFarmOutput as ReorderFarmOutput

from .population import PopulationStates as PopulationStates
from .population import PopulationModel as PopulationModel



import numpy as np
import pandas as pd
from os import PathLike
from xarray import Dataset, open_dataset

from foxes.data import STATES
from foxes.utils import PandasFileHelper, weibull_weights
from foxes.config import config, get_input_path
import foxes.variables as FV
import foxes.constants as FC

from .states_table import StatesTable


class WeibullSectors(StatesTable):
    """
    States with wind speed from Weibull parameters
    from a NetCDF file

    Attributes
    ----------
    ws_bins: numpy.ndarray
        The wind speed bins, including
        lower and upper bounds, shape: (n_ws_bins+1,)
    var2ncvar: dict
        Mapping from variable names to variable names
        in the nc file
    sel: dict
        Subset selection via xr.Dataset.sel()
    isel: dict
        Subset selection via xr.Dataset.isel()
    rpars: dict
        Additional parameters for reading the file
    RDICT: dict
        Default xarray file reading parameters

    :group: input.states

    """

    RDICT = {}

    def __init__(
        self,
        data_source,
        output_vars,
        ws_bins=None,
        var2ncvar={},
        sel=None,
        isel=None,
        read_pars={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or xarray.Dataset or pandas.DataFrame
            Either path to NetCDF or csv file or data
        output_vars: list of str
            The output variables
        ws_bins: list of float, optional
            The wind speed bins, including
            lower and upper bounds
        var2ncvar: dict
            Mapping from variable names to variable names
            in the nc file
        sel: dict, optional
            Subset selection via xr.Dataset.sel()
        isel: dict, optional
            Subset selection via xr.Dataset.isel()
        read_pars: dict
            Additional parameters for reading the file
        kwargs: dict, optional
            Additional arguments for the base class

        """
        super().__init__(data_source, output_vars, var2col={}, **kwargs)
        self.ws_bins = None if ws_bins is None else np.asarray(ws_bins)
        self.var2ncvar = var2ncvar
        self.sel = sel if sel is not None else {}
        self.isel = isel if isel is not None else {}
        self.rpars = read_pars

        if FV.WS not in self.ovars:
            raise ValueError(
                f"States '{self.name}': Expecting output variable '{FV.WS}', got {self.ovars}"
            )
        for v in [FV.WEIBULL_A, FV.WEIBULL_k, FV.WEIGHT]:
            if v in self.ovars:
                raise ValueError(
                    f"States '{self.name}': Cannot have '{v}' as output variable"
                )

        self._original_data = None

    def __repr__(self):
        return f"{type(self).__name__}(n_wd={self._n_wd}, n_ws={self._n_ws})"

    def _read_data(self, algo, point_coord=None, verbosity=0):
        """
        Extracts data from file or Dataset.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        point_coord: str, optional
            The coordinate name representing the point index
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        data: xarray.Dataset
            The input data

        """
        # store original data
        if self._original_data is not None:
            self._data = self._original_data
            self._original_data = None

        # read file or grab data
        cwd = self.var2ncvar.get(FV.WD, FV.WD)
        if isinstance(self.data_source, (str, PathLike)):
            fpath = get_input_path(self.data_source)
            if not fpath.is_file():
                if verbosity > 0:
                    print(
                        f"States '{self.name}': Reading static data '{fpath}' from context '{STATES}'"
                    )
                fpath = algo.dbook.get_file_path(STATES, fpath.name, check_raw=False)
                if verbosity > 0:
                    print(f"Path: {fpath}")
            elif verbosity > 0:
                print(f"States '{self.name}': Reading file {fpath}")
            rpars = dict(self.RDICT, **self.rpars)
            if fpath.suffix == ".nc":
                data = open_dataset(fpath, engine=config.nc_engine, **rpars)
            else:
                data = PandasFileHelper().read_file(fpath, **rpars)
                data.index.name = cwd
                data = data.to_xarray()
            self._original_data = data

        elif isinstance(self.data_source, Dataset):
            data = self.data_source

        elif isinstance(self.data_source, pd.DataFrame):
            data = self.data_source
            data.index.name = cwd
            data = data.to_xarray()

        # optionally select a subset
        if self.isel is not None and len(self.isel):
            data = data.isel(**self.isel)
        if self.sel is not None and len(self.sel):
            data = data.sel(**self.sel)

        # remove wd 360 from the end, if wd 0 is given:
        wd = data[cwd].to_numpy()
        if wd[0] == 0.0 and wd[-1] == 360.0:
            data = data.isel({cwd: np.s_[:-1]})

        # construct wind speed bins and bin deltas
        cws = self.var2ncvar.get(FV.WS, FV.WS)
        if self.ws_bins is not None:
            wsb = self.ws_bins
            wss = 0.5 * (wsb[:-1] + wsb[1:])
        elif cws in data:
            wss = data[cws].to_numpy()
            wsb = np.zeros((len(wss) + 1,), dtype=config.dtype_double)
            wsb[1:-1] = 0.5 * (wss[1:] + wss[:-1])
            wsb[0] = wss[0] - 0.5 * wsb[1]
            wsb[-1] = wss[-1] + 0.5 * wsb[-2]
            self.ws_bins = wsb
        else:
            raise ValueError(
                f"States '{self.name}': Expecting ws_bins argument, since '{cws}' not found in data"
            )
        wsd = wsb[1:] - wsb[:-1]
        n_ws = len(wss)
        del wsb

        # prepare data binning
        if self._original_data is None:
            self._original_data = self.data_source
        cpt = (
            self.var2ncvar.get(FC.POINT, FC.POINT) if point_coord is not None else None
        )
        n_wd = data.sizes[cwd]
        self.BIN_WD = self.var("bin_wd")
        self.BIN_WS = self.var("bin_ws")
        self.POINT = self.var("point")
        if cpt is None:
            n_pt = 0
            shp = (n_wd, n_ws)
            dms = (self.BIN_WD, self.BIN_WS)
        else:
            n_pt = data.sizes[cpt]
            shp = (n_wd, n_ws, n_pt)
            dms = (self.BIN_WD, self.BIN_WS, self.POINT)

        # create binned data
        self._data = {
            FV.WD: np.zeros(shp, dtype=config.dtype_double),
            FV.WS: np.zeros(shp, dtype=config.dtype_double),
        }
        if cpt is None:
            self._data[FV.WD][:] = data[cwd].to_numpy()[:, None]
            self._data[FV.WS][:] = wss[None, :]
        else:
            self._data[FV.WD][:] = data[cwd].to_numpy()[:, None, None]
            self._data[FV.WS][:] = wss[None, :, None]
        for v in [FV.WEIBULL_A, FV.WEIBULL_k, FV.WEIGHT] + self.ovars:
            if v not in [FV.WS, FV.WD] and v not in self.fixed_vars:
                w = self.var2ncvar.get(v, v)
                if w not in data:
                    raise KeyError(
                        f"States '{self.name}': Missing variable '{w}' in data, found {list(data.data_vars.keys())}"
                    )
                d = data[w]
                iws = d.dims.index(cws) if cws in d.dims else -1
                iwd = d.dims.index(cwd) if cwd in d.dims else -1
                ipt = d.dims.index(cpt) if cpt is not None and cpt in d.dims else -1
                d = d.to_numpy()
                if v in [FV.WEIBULL_A, FV.WEIBULL_k, FV.WEIGHT]:
                    if cws in data[w].dims:
                        raise ValueError(
                            f"States '{self.name}': Cannot have '{cws}' as dimension in variable '{v}', got {data[w].dims}"
                        )
                    if cwd not in data[w].dims:
                        raise ValueError(
                            f"States '{self.name}': Expecting '{cwd}' as dimension in variable '{v}', got {data[w].dims}"
                        )
                if iws < 0 and iwd < 0 and ipt < 0:
                    self.fixed_vars[v] = d.to_numpy()
                elif iws >= 0 and iwd >= 0 and ipt >= 0:
                    self._data[v] = np.moveaxis(d, [iwd, iws, ipt], [0, 1, 2])
                elif iws >= 0 and iwd >= 0 and ipt < 0:
                    if cpt is None:
                        self._data[v] = np.moveaxis(d, [iwd, iws], [0, 1])
                    else:
                        self._data[v] = np.zeros(shp, dtype=config.dtype_double)
                        self._data[v][:] = np.moveaxis(d, [iwd, iws], [0, 1])[
                            :, :, None
                        ]
                elif iws >= 0 and iwd < 0 and ipt >= 0:
                    self._data[v] = np.zeros(shp, dtype=config.dtype_double)
                    self._data[v][:] = np.moveaxis(d, [iws, ipt], [0, 1])[None, :, :]
                elif iws < 0 and iwd >= 0 and ipt >= 0:
                    self._data[v] = np.zeros(shp, dtype=config.dtype_double)
                    self._data[v][:] = np.moveaxis(d, [iwd, ipt], [0, 1])[:, None, :]
                elif iws >= 0 and iwd < 0 and ipt < 0:
                    self._data[v] = np.zeros(shp, dtype=config.dtype_double)
                    if cpt is None:
                        self._data[v][:] = d[None, :]
                    else:
                        self._data[v][:] = d[None, :, None]
                elif iws < 0 and iwd >= 0 and ipt < 0:
                    self._data[v] = np.zeros(shp, dtype=config.dtype_double)
                    if cpt is None:
                        self._data[v][:] = d[:, None]
                    else:
                        self._data[v][:] = d[:, None, None]
                elif iws < 0 and iwd < 0 and ipt >= 0:
                    self._data[v] = np.zeros(shp, dtype=config.dtype_double)
                    self._data[v][:] = d[None, None, :]

        # compute Weibull weights
        self._data[FV.WEIGHT] *= weibull_weights(
            ws=wss[None, :, None] if cpt is not None else wss[None, :],
            ws_deltas=wsd[None, :, None] if cpt is not None else wsd[None, :],
            A=self._data.pop(FV.WEIBULL_A),
            k=self._data.pop(FV.WEIBULL_k),
        )

        # translate binned data to states
        self._n_pt = n_pt
        self._n_wd = n_wd
        self._n_ws = n_ws
        self._N = n_wd * n_ws
        self._data = Dataset(
            data_vars={v: (dms, d) for v, d in self._data.items()},
        )

        return data

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        self._read_data(algo, verbosity=0)

        tmp = {}
        for v, d in self._data.data_vars.items():
            if self.POINT in d.dims:
                raise TypeError(
                    f"States '{self.name}': Variable '{v}' has unsupported dimension '{self.POINT}', dims = {d.dims}"
                )
            else:
                tmp[v] = d.to_numpy().reshape(self._N)
        self._data = tmp
        del tmp

        self._data = pd.DataFrame(data=self._data, index=np.arange(self._N))
        self._data.index.name = FC.STATE

        return super().load_data(algo, verbosity)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from foxes.input.states import StatesTable
from foxes.core import WindFarm, Turbine
from foxes.algorithms import Downwind
from foxes.models.turbine_models import SetFarmVars
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC

from .output import Output


class TurbineTypeCurves(Output):
    """
    Creates power and ct curves for turbine
    types, optionally including derating/boost.

    Attributes
    ----------
    mbook: foxes.models.ModelBook
        The model book

    :group: output

    """

    def __init__(self, mbook, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        mbook: foxes.models.ModelBook
            The model book
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.mbook = mbook

    def plot_curves(
        self,
        turbine_type,
        variables,
        P_max=None,
        titles=None,
        x_label=None,
        y_labels=None,
        ws_min=0.0,
        ws_max=30.0,
        ws_step=0.1,
        ti=0.05,
        rho=1.225,
        axs=None,
        figsize=None,
        pmax_args={},
        **kwargs,
    ):
        """
        Plot the power or ct curve.

        Parameters
        ----------
        turbine_type: str
            The turbine type name from the
            model book
        variables: str or list of str
            For example FV.P or FV.CT
        P_max: float, optional
            The power mask value, if of interest
        titles: list of str, optional
            The plot titles, one for each variable
        x_label: str, optional
            The x axis label
        y_labels: list of str, optional
            The y axis lables, one for each variable
        ws_min: float
            The minimal wind speed
        ws_max: float
            The maximal wind speed
        ws_step: float
            The wind speed step size
        ti: float
            The TI value
        rho: float
            The air density value
        axs: list of pyplot.Axis, optional
            The axis, one for each variable
        figsize: tuple
            The figsize argument for plt.subplots()
            in case ax is not provided
        pmax_args: dict, optional
            Additionals parameters for plt.plot()
            for power mask case
        kwargs: dict, optional
            Additional parameters for plt.plot()

        Returns
        -------
        axs: list of pyplot.Axis
            The plot axes, one for each variable

        """
        if self.nofig:
            return None

        vars = [variables] if isinstance(variables, str) else variables
        if isinstance(titles, str):
            titles = [titles]
        elif titles is None:
            titles = [None for v in vars]
        if isinstance(y_labels, str):
            y_labels = [y_labels]
        elif y_labels is None:
            y_labels = [None for v in vars]
        if not isinstance(axs, (list, tuple, np.ndarray)):
            axs = [axs]

        ws = np.arange(ws_min, ws_max + ws_step, ws_step, dtype=config.dtype_double)
        n_states = len(ws)
        sdata = pd.DataFrame(index=range(n_states))
        sdata.index.name = FC.STATE
        sdata[FV.WS] = ws

        models = [turbine_type]

        states = StatesTable(
            sdata,
            output_vars={FV.WS, FV.WD, FV.TI, FV.RHO},
            fixed_vars={FV.WD: 270.0, FV.TI: ti, FV.RHO: rho},
        )

        farm = WindFarm()
        farm.add_turbine(
            Turbine(xy=[0.0, 0.0], turbine_models=models),
            verbosity=0,
        )

        algo = Downwind(farm, states, wake_models=[], mbook=self.mbook, verbosity=0)

        results = algo.calc_farm()

        if P_max is not None:
            sname = f"_{type(self).__name__}_set_Pmax"
            self.mbook.turbine_models[sname] = SetFarmVars()
            self.mbook.turbine_models[sname].add_var(FV.MAX_P, P_max)
            models += [sname, "PMask"]

            farm = WindFarm()
            farm.add_turbine(
                Turbine(xy=[0.0, 0.0], turbine_models=models),
                verbosity=0,
            )

            algo = Downwind(farm, states, wake_models=[], mbook=self.mbook, verbosity=0)

            results1 = algo.calc_farm()

            del self.mbook.turbine_models[sname]

        for i, v in enumerate(vars):
            ax = axs[i]
            if ax is None:
                __, ax = plt.subplots(figsize=figsize)

            pargs = {"linewidth": 2.5}
            pargs.update(kwargs)
            ax.plot(ws, results[v][:, 0], label="default", **pargs)
            if P_max is not None:
                pargs = {"linewidth": 1.8}
                pargs.update(pmax_args)
                ax.plot(ws, results1[v][:, 0], label="PMask", **pargs)

            if v == FV.P:
                vv = "Power curve"
            elif v == FV.CT:
                vv = "Thrust curve"
            else:
                vv = v
            t = f"{vv}, {turbine_type}" if titles[i] is None else titles[i]
            ax.set_title(t)

            lb = "Wind speed [m/s]" if x_label is None else x_label
            ax.set_xlabel(lb)

            if y_labels[i] is None:
                if v == FV.P:
                    lb = "Power [kW]"
                elif v == FV.CT:
                    lb = "ct [-]"
                else:
                    lb = v
            else:
                lb = y_labels[i]
            ax.set_ylabel(lb)

            ax.grid()

            if P_max is not None:
                ax.legend()

        return axs



import time
import argparse
import numpy as np
from xarray import Dataset

import foxes
import foxes.variables as FV
import foxes.constants as FC

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2016_linear"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument("-e", "--engine", help="The engine", default=None)
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    states = foxes.input.states.Timeseries(
        data_source="timeseries_3000.csv.gz",
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=np.array([500.0, 500.0]),
        step_vectors=np.array([[500.0, 0], [0, 700.0]]),
        steps=(5, 5),
        turbine_models=args.tmodels + ["NREL5MW"],
    )

    k_values = np.linspace(0.01, 0.1, 10)
    n_pop = len(k_values)
    pop_data = np.zeros((n_pop, farm.n_turbines))
    pop_data[:] = k_values[:, None]
    pop_data = Dataset(
        {
            FV.K: (("i", FC.TURBINE), pop_data),
        }
    )
    print("\nInput population data:\n\n", pop_data)
    print(f"\n{FV.K} values: {pop_data[FV.K].values[:, 0]}\n")

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        wake_models=args.wakes,
        rotor_model=args.rotor,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        population_params=dict(
            data_source=pop_data,
            index_coord="i",
            turbine_coord=FC.TURBINE,
        ),
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()
        print("Calc time =", time1 - time0, "\n")

    pop_results = algo.population_model.farm2pop_results(algo, farm_results)
    print("\nPopulation results:\n\n", pop_results)
    print(f"\n{FV.K} values: {pop_results[FV.K].values[:, 0, 0]}")

    P_mean = (
        (pop_results[FV.P] * pop_results[FV.WEIGHT])
        .sum(dim=FC.STATE)
        .sum(dim=FC.TURBINE)
    )
    print("\nFarm P_mean:", P_mean.values)



import numpy as np

from foxes.config import config
from foxes.core import TurbineInductionModel
import foxes.variables as FV
import foxes.constants as FC


class SelfSimilar(TurbineInductionModel):
    """
    The self-similar induction wake model
    from Troldborg and Meyer Forsting

    Notes
    -----
    References:
    [1] Troldborg, Niels, and Alexander Raul Meyer Forsting.
    "A simple model of the wind turbine induction zone derived
    from numerical simulations."
    Wind Energy 20.12 (2017): 2011-2020.
    https://onlinelibrary.wiley.com/doi/full/10.1002/we.2137

    [2] Forsting, Alexander R. Meyer, et al.
    "On the accuracy of predicting wind-farm blockage."
    Renewable Energy (2023).
    https://www.sciencedirect.com/science/article/pii/S0960148123007620

    Attributes
    ----------
    alpha: float
        The alpha parameter
    beta: float
        The beta parameter
    gamma: float
        The gamma parameter
    pre_rotor_only: bool
        Calculate only the pre-rotor region
    induction: foxes.core.AxialInductionModel or str
        The induction model

    :group: models.wake_models.induction

    """

    def __init__(
        self,
        superposition="ws_linear",
        induction="Madsen",
        alpha=8 / 9,
        beta=np.sqrt(2),
        gamma=1.1,
        pre_rotor_only=False,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The wind speed superposition.
        induction: foxes.core.AxialInductionModel or str
            The induction model
        alpha: float
            The alpha parameter
        beta: float
            The beta parameter
        gamma: float
            The gamma parameter
        pre_rotor_only: bool
            Calculate only the pre-rotor region

        """
        super().__init__(wind_superposition=superposition)
        self.induction = induction
        self.pre_rotor_only = pre_rotor_only
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        return f"{type(self).__name__}({self.wind_superposition}, induction={iname}, gamma={self.gamma})"

    @property
    def affects_ws(self):
        """
        Flag for wind speed wake models

        Returns
        -------
        dws: bool
            If True, this model affects wind speed

        """
        return True

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return super().sub_models() + [self.induction]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if isinstance(self.induction, str):
            self.induction = algo.mbook.axial_induction[self.induction]
        super().initialize(algo, verbosity, force)

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        if self.has_uv:
            duv = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints, 2),
                dtype=config.dtype_double,
            )
            return {FV.UV: duv}
        else:
            dws = np.zeros(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                dtype=config.dtype_double,
            )
            return {FV.WS: dws}

    def _mu(self, x_R):
        """Helper function: define mu (eqn 11 from [1])"""
        return 1 + (x_R / np.sqrt(1 + x_R**2))

    def _a0(self, ct, x_R):
        """Helper function: define a0 with gamma factor, eqn 8 from [2]"""
        return self.induction.ct2a(self.gamma * ct)

    def _a(self, ct, x_R):
        """Helper function: define axial shape function (eqn 11 from [1])"""
        return self._a0(ct, x_R) * self._mu(x_R)

    def _r_half(self, x_R):
        """Helper function: using eqn 13 from [2]"""
        return np.sqrt(0.587 * (1.32 + x_R**2))

    def _rad_fn(self, x_R, r_R):
        """Helper function: define radial shape function (eqn 12 from [1])"""
        with np.errstate(over="ignore"):
            result = (
                1 / np.cosh(self.beta * (r_R) / self._r_half(x_R))
            ) ** self.alpha  # * (x_R < 0)
        return result

    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        # get ct
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
        )

        # get R
        R = 0.5 * self.get_data(
            FV.D,
            FC.STATE_TARGET_TPOINT,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=False,
            downwind_index=downwind_index,
        )

        # get x, r and R etc. Rounding for safe x < 0 condition below
        x_R = np.round(wake_coos[..., 0] / R, 12)
        r_R = np.linalg.norm(wake_coos[..., 1:3], axis=-1) / R

        def add_wake(sp_sel, wake_deltas, blockage):
            """adds to wake deltas"""
            if self.has_uv:
                assert self.has_vector_wind_superp, (
                    f"Wake model {self.name}: Missing vector wind superposition, got '{self.wind_superposition}'"
                )
                wdeltas = {FV.WS: blockage}
                self.vec_superp.wdeltas_ws2uv(
                    algo, fdata, tdata, downwind_index, wdeltas, sp_sel
                )
                wake_deltas[FV.UV] = self.vec_superp.add_wake_vector(
                    algo,
                    mdata,
                    fdata,
                    tdata,
                    downwind_index,
                    sp_sel,
                    wake_deltas[FV.UV],
                    wdeltas.pop(FV.UV),
                )
            else:
                self.superp[FV.WS].add_wake(
                    algo,
                    mdata,
                    fdata,
                    tdata,
                    downwind_index,
                    sp_sel,
                    FV.WS,
                    wake_deltas[FV.WS],
                    blockage,
                )

        # select values
        sp_sel = (ct > 1e-8) & (x_R <= 0)  # upstream
        if np.any(sp_sel):
            # velocity eqn 10 from [1]
            xr = x_R[sp_sel]
            blockage = self._a(ct[sp_sel], xr) * self._rad_fn(xr, r_R[sp_sel])

            add_wake(sp_sel, wake_deltas, -blockage)

        # set area behind to mirrored value EXCEPT for area behind turbine
        if not self.pre_rotor_only:
            sp_sel = (ct > 1e-8) & (x_R > 0) & (r_R > 1)
            if np.any(sp_sel):
                # velocity eqn 10 from [1]
                xr = x_R[sp_sel]
                blockage = self._a(ct[sp_sel], -xr) * self._rad_fn(-xr, r_R[sp_sel])

                add_wake(sp_sel, wake_deltas, blockage)

        return wake_deltas



from foxes.core.wake_deflection import WakeDeflection


class NoDeflection(WakeDeflection):
    """
    Switch of wake deflection

    :group: models.wake_deflections

    """

    def calc_deflection(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        coos,
    ):
        """
        Calculates the wake deflection.

        This function optionally adds FC.WDEFL_ROT_ANGLE or
        FC.WDEFL_DWS_FACTOR to the tdata.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        Returns
        -------
        coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        return coos

    def get_yaw_alpha_seq(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Computes sequential wind vector rotation angles.

        Wind vector rotation angles are computed at the
        current trace points due to a yawed rotor
        for sequential runs.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        x: numpy.ndarray
            The distance from the wake causing rotor
            for the first n_times subsequent time steps,
            shape: (n_times,)

        Returns
        -------
        alpha: numpy.ndarray
            The delta WD result at the x locations,
            shape: (n_times,)

        """
        return None



from pathlib import Path
import inspect

import foxes
import foxes.variables as FV


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    tfile = thisdir / "NREL-5MW-D126-H90.csv"
    sfile = thisdir / "states.csv.gz"
    lfile = thisdir / "test_farm.csv"
    cases = [
        ("grid400", "rotor_points", None),
        ("grid4", "rotor_points", 0.15),
        ("grid9", "rotor_points", 0.07),
        ("centre", "axiwake5", 0.03),
        ("centre", "axiwake10", 0.0081),
        ("centre", "grid9", 0.07),
        ("centre", "grid16", 0.05),
        ("centre", "grid36", 0.025),
    ]

    base_results = None
    with foxes.Engine.new("threads", chunk_size_states=100):
        for rotor, pwake, lim in cases:
            print(f"\nENTERING CASE {(rotor, pwake, lim)}\n")

            mbook = foxes.models.ModelBook()
            ttype = foxes.models.turbine_types.PCtFile(
                data_source=tfile, var_ws_ct=FV.REWS, var_ws_P=FV.REWS
            )
            mbook.turbine_types[ttype.name] = ttype

            states = foxes.input.states.StatesTable(
                data_source=sfile,
                output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
                var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti"},
                fixed_vars={FV.RHO: 1.225},
            )

            farm = foxes.WindFarm()
            foxes.input.farm_layout.add_from_file(
                farm, lfile, turbine_models=[ttype.name], verbosity=1
            )

            algo = foxes.algorithms.Downwind(
                farm,
                states,
                mbook=mbook,
                rotor_model=rotor,
                wake_models=["Bastankhah025_linear_k002"],
                wake_frame="rotor_wd",
                partial_wakes=pwake,
                verbosity=1,
            )

            data = algo.calc_farm()

            df = data.to_dataframe()[
                [FV.AMB_WD, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]
            ]

            print()
            print("TRESULTS\n")
            print(df)

            df = df.reset_index()

            if base_results is None:
                base_results = df

            else:
                print(f"CASE {(rotor, pwake, lim)}")
                delta = df - base_results
                print(delta)
                print(delta.min(), delta.max())
                chk = delta[FV.REWS].abs()
                print(f"CASE {(rotor, pwake, lim)}:", chk.max())

                assert (chk < lim).all()


if __name__ == "__main__":
    test()



import numpy as np
from xarray import Dataset

from foxes.algorithms import Iterative
from foxes.config import config
from foxes.core import get_engine
import foxes.variables as FV
import foxes.constants as FC

from . import models as mdls


class Sequential(Iterative):
    """
    A sequential calculation of states without chunking.

    This is of use for the evaluation in simulation
    environments that do not support multi-state computations,
    like FMUs.

    Attributes
    ----------
    ambient: bool
        Flag for ambient calculation
    calc_pars: dict
        Parameters for model calculation.
        Key: model name str, value: parameter dict
    states0: foxes.core.States
        The original states
    points: numpy.ndarray
        The points of interest, shape: (n_states, n_points, 3)
    plugins: list of foxes.algorithm.sequential.SequentialIterPlugin
        The plugins, updated with every iteration
    outputs: list of str
        The output variables
    :group: algorithms.sequential

    """

    @classmethod
    def get_model(cls, name):
        """
        Get the algorithm specific model

        Parameters
        ----------
        name: str
            The model name

        Returns
        -------
        model: foxes.core.model
            The model

        """
        try:
            return getattr(mdls, name)
        except AttributeError:
            return super().get_model(name)

    def __init__(
        self,
        farm,
        states,
        *args,
        points=None,
        ambient=False,
        calc_pars={},
        plugins=[],
        outputs=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        farm: foxes.WindFarm
            The wind farm
        states: foxes.core.States
            The ambient states
        args: tuple, optional
            Additional arguments for Downwind
        points: numpy.ndarray, optional
            The points of interest, shape: (n_states, n_points, 3)
        ambient: bool
            Flag for ambient calculation
        calc_pars: dict
            Parameters for model calculation.
            Key: model name str, value: parameter dict
        plugins: list of foxes.algorithm.sequential.SequentialIterPlugin
            The plugins, updated with every iteration
        outputs: list of str, optional
            The output variables
        kwargs: dict, optional
            Additional arguments for Downwind

        """
        super().__init__(farm, mdls.SeqState(states), *args, **kwargs)
        self.ambient = ambient
        self.calc_pars = calc_pars
        self.states0 = self.states.states
        self.points = points
        self.plugins = plugins
        self.outputs = outputs if outputs is not None else self.DEFAULT_FARM_OUTPUTS

        self._verbo0 = self.verbosity + 1
        self.verbosity -= 1
        get_engine().verbosity -= 2

        self._i = None

    @property
    def iterating(self):
        """
        Flag for running iteration

        Returns
        -------
        itr: bool
            True if currently iterating

        """
        return self._i is not None

    def get_models_data(self, sel=None, isel=None):
        if sel is not None and len(sel):
            raise ValueError(f"calc_points does not support sel, got sel={sel}")
        if isel is not None and len(isel):
            raise ValueError(f"calc_points does not support isel, got isel={isel}")
        return self._model_data.isel({FC.STATE: [self.counter]})

    def __iter__(self):
        """Initialize the iterator"""

        if not self.iterating:
            if not self.initialized:
                self.initialize()
            self.print_deco("calc_farm")

            self._inds = self.states0.index()
            self._i = 0
            self._counter = 0

            self._it = 0
            mlist, __ = self._collect_farm_models(
                None, self.calc_pars, ambient=self.ambient
            )
            self._calc_farm_vars(mlist)
            self._it = None

            self._model_data = Dataset(**super().get_models_idata())

            if self._verbo0 > 0:
                print("\nInput data:\n")
                print(self._model_data)
                print("\nOutput farm variables:", ", ".join(self.farm_vars))
                print()

            sts = self._model_data[FC.STATE].to_numpy()
            self._farm_results = Dataset(
                coords={FC.STATE: sts},
                data_vars={
                    v: (
                        (FC.STATE, FC.TURBINE),
                        np.zeros(
                            (len(sts), self.n_turbines), dtype=config.dtype_double
                        ),
                    )
                    for v in self.farm_vars
                },
            )
            self._farm_results[FC.TNAME] = ((FC.TURBINE,), self.farm.turbine_names)
            if FV.ORDER in self._farm_results:
                self._farm_results[FV.ORDER] = self._farm_results[FV.ORDER].astype(
                    config.dtype_int
                )
            self._farm_results_dwnd = self._farm_results.copy(deep=True)

            self._point_results = None

            for p in self.plugins:
                p.initialize(self)

        return self

    def __next__(self):
        """Run calculation for current step, then iterate to next"""

        if self._i < len(self._inds):
            self._counter = self._i
            self.states._counter = self._i
            self.states._size = 1
            self.states._indx = self._inds[self._i]

            if self._verbo0 > 0:
                print(f"{self.name}: Running state {self.states.index()[0]}")

            self.reset_chunk_store()
            fres, fres_dnwnd = super().calc_farm(
                outputs=self.farm_vars,
                finalize=False,
                ret_dwnd_order=True,
                **self.calc_pars,
            )

            for v in self._farm_results.data_vars.keys():
                if FC.STATE in self._farm_results[v].dims:
                    self._farm_results[v].loc[{FC.STATE: [self.index]}] = fres[v]
                    self._farm_results_dwnd[v].loc[{FC.STATE: [self.index]}] = (
                        fres_dnwnd[v]
                    )

            if self.points is None:
                for p in self.plugins:
                    p.update(self, fres)

                self._i += 1
                return fres

            else:
                pres = super().calc_points(fres, points=self.points, finalize=False)

                if self._point_results is None:
                    n_states = self._model_data.sizes[FC.STATE]
                    self._point_results = Dataset(
                        coords={
                            FC.STATE: self._model_data[FC.STATE].to_numpy(),
                            **{c: d for c, d in pres.coords.items() if c != FC.STATE},
                        },
                        data_vars={
                            v: (
                                d.dims,
                                np.zeros([n_states] + list(d.shape[1:]), dtype=d.dtype),
                            )
                            for v, d in pres.data_vars.items()
                            if d.dims[0] == FC.STATE
                        },
                    )
                    for v, d in pres.data_vars.items():
                        if FC.STATE not in d.dims:
                            self._point_results[v] = d

                for v in self._point_results.data_vars.keys():
                    if FC.STATE in self._point_results[v].dims:
                        self._point_results[v].loc[{FC.STATE: [self.index]}] = pres[v]

                for p in self.plugins:
                    p.update(self, fres, pres)

                self._i += 1
                return fres, pres

        else:
            del self._model_data

            self._i = None
            self.states._counter = None
            self.states._size = len(self._inds)
            self.states._indx = self._inds

            for p in self.plugins:
                p.finalize(self)

            raise StopIteration

    @property
    def size(self):
        """
        The total number of iteration steps

        Returns
        -------
        s: int
            The total number of iteration steps

        """
        return self.states.size()

    @property
    def counter(self):
        """
        The current index counter

        Returns
        -------
        i: int
            The current index counter

        """
        return self._counter if self.iterating else None

    @property
    def index(self):
        """
        The current index

        Returns
        -------
        indx: int
            The current index

        """
        return self.states._indx if self.iterating else None

    def states_i0(self, counter, algo=None):
        """
        Returns counter or index

        Parameters
        ----------
        counter: bool
            Flag for counter
        algo: object, optional
            Dummy argument, due to consistency with
            foxes.core.Data.states_i0

        Returns
        -------
        i0: int
            The counter or index

        """
        return self.counter if counter else self.index

    @property
    def farm_results(self):
        """
        The overall farm results

        Returns
        -------
        results: xarray.Dataset
            The overall farm results

        """
        return self._farm_results

    @property
    def farm_results_downwind(self):
        """
        The overall farm results, with turbine
        dimension in downwind order

        Returns
        -------
        results: xarray.Dataset
            The overall farm results

        """
        return self._farm_results_dwnd

    @property
    def cur_farm_results(self):
        """
        The current farm results

        Returns
        -------
        results: xarray.Dataset
            The current farm results

        """
        return self._farm_results.isel({FC.STATE: [self.counter]})

    @property
    def point_results(self):
        """
        The overall point results

        Returns
        -------
        results: xarray.Dataset
            The overall point results

        """
        return self._point_results

    @property
    def cur_point_results(self):
        """
        The current point results

        Returns
        -------
        results: xarray.Dataset
            The current point results

        """
        return self._point_results.isel({FC.STATE: [self.counter]})

    def calc_farm(self):
        """
        Calculate farm data.

        Returns
        -------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)

        """
        if not self.iterating:
            raise ValueError("calc_farm call is only allowed during iterations")
        return self.cur_farm_results

    def calc_points(
        self,
        farm_results,
        points,
        **kwargs,
    ):
        """
        Calculate data at a given set of points.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)
        points: numpy.ndarray
            The points of interest, shape: (n_states, n_points, 3)
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices

        Returns
        -------
        point_results: xarray.Dataset
            The point results. The calculated variables have
            dimensions (state, point)

        """
        if not self.iterating:
            raise ValueError("calc_points call is only allowed during iterations")

        return super().calc_points(farm_results, points, finalize=False, **kwargs)



from abc import abstractmethod

from .model import Model
from foxes.utils import new_instance


class VerticalProfile(Model):
    """
    Abstract base class for vertical profiles.

    :group: core

    """

    @abstractmethod
    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        return []

    @abstractmethod
    def calculate(self, tdata, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        tdata: dict
            The target point data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """
        pass

    @classmethod
    def new(cls, profile_type, *args, **kwargs):
        """
        Run-time vertical profile factory.

        Parameters
        ----------
        profile_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for the constructor
        kwargs: dict, optional
            Additional parameters for the constructor

        """
        return new_instance(cls, profile_type, *args, **kwargs)



import numpy as np

from foxes.core.wake_deflection import WakeDeflection
from foxes.algorithms import Sequential
import foxes.constants as FC
import foxes.variables as FV


class JimenezDeflection(WakeDeflection):
    """
    Yawed rotor wake defection according to the Jimenez model

    Notes
    -----
    Reference:
    Jiménez, Á., Crespo, A. and Migoya, E. (2010), Application of a LES technique to characterize
    the wake deflection of a wind turbine in yaw. Wind Energ., 13: 559-572. doi:10.1002/we.380


    Attributes
    ----------
    rotate: bool
        If True, rotate local wind vector at evaluation points.
        If False, multiply wind speed with cos(angle) instead.
        If None, do not modify the wind vector, only the path.
    beta: float
        The beta coefficient of the Jimenez model
    step_x: float
        The x step in m for integration

    :group: models.wake_deflections

    """

    def __init__(self, rotate=True, beta=0.1, step_x=10.0):
        """
        Constructor.

        Parameters
        ----------
        rotate: bool, optional
            If True, rotate local wind vector at evaluation points.
            If False, multiply wind speed with cos(angle) instead.
            If None, do not modify the wind vector, only the path.
        beta: float
            The beta coefficient of the Jimenez model
        step_x: float
            The x step in m for integration

        """
        super().__init__()
        self.rotate = rotate
        self.beta = beta
        self.step_x = step_x

    def __repr__(self):
        s = f"{type(self).__name__}("
        s += f"rotate={self.rotate}, beta={self.beta}, step_x={self.step_x}"
        s += ")"
        return s

    @property
    def has_uv(self):
        """
        This model uses wind vector data

        Returns
        -------
        hasuv: bool
            Flag for wind vector data

        """
        return self.rotate is not None and self.rotate

    def calc_deflection(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        coos,
    ):
        """
        Calculates the wake deflection.

        This function optionally adds FC.WDEFL_ROT_ANGLE or
        FC.WDEFL_DWS_FACTOR to the tdata.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        Returns
        -------
        coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """

        if FV.YAWM not in fdata:
            return coos

        # take rotor average:
        xyz = np.einsum("stpd,p->std", coos, tdata[FC.TWEIGHTS])
        x = xyz[:, :, 0]
        y = xyz[:, :, 1]
        z = xyz[:, :, 2]

        # get ct:
        ct = self.get_data(
            FV.CT,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        # get gamma:
        gamma = self.get_data(
            FV.YAWM,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )

        sel = (x > 1e-8) & (x < 1e10) & (ct > 1e-8) & (np.abs(gamma) > 1e-8)
        delwd = np.zeros_like(coos[..., 0])
        n_sel = np.sum(sel)
        if n_sel > 0:
            # apply selection:
            gamma = np.deg2rad(gamma[sel])
            ct = ct[sel]
            x = x[sel]

            # get rotor diameter:
            D = self.get_data(
                FV.D,
                FC.STATE_TARGET,
                lookup="w",
                algo=algo,
                fdata=fdata,
                tdata=tdata,
                downwind_index=downwind_index,
                upcast=True,
                selection=sel,
            )[:, None]

            # define x path:
            xmax = np.max(x)
            n_x = int(xmax / self.step_x)
            if xmax > n_x * self.step_x:
                n_x += 1
            delx = np.arange(n_x + 1) * self.step_x
            delx = np.minimum(delx[None, :], x[:, None])
            dx = delx[:, 1:] - delx[:, :-1]
            delx = delx[:, :-1]

            # integrate deflection of y along the x path:
            alpha0 = (
                -(np.cos(gamma[:, None]) ** 2)
                * np.sin(gamma[:, None])
                * ct[:, None]
                / 2
            )
            y[sel] += np.sum(
                np.tan(alpha0 / (1 + self.beta * delx / D) ** 2) * dx, axis=-1
            )
            del delx, dx
            coos[..., 1] = y[:, :, None]

            # calculate wind vector modification at evaluation points:
            if self.rotate is not None:
                # delta wd at evaluation points, if within wake radius:
                r2 = (y[sel, None] ** 2 + z[sel, None] ** 2) / D**2
                WD2 = (1 + self.beta * x[:, None] / D) ** 2
                delwd[sel] = np.where(r2 <= WD2 / 4, alpha0 / WD2, 0)

                if self.rotate:
                    tdata[FC.WDEFL_ROT_ANGLE] = np.rad2deg(delwd)
                else:
                    tdata[FC.WDEFL_DWS_FACTOR] = np.cos(delwd)

        return coos

    def get_yaw_alpha_seq(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Computes sequential wind vector rotation angles.

        Wind vector rotation angles are computed at the
        current trace points due to a yawed rotor
        for sequential runs.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        x: numpy.ndarray
            The distance from the wake causing rotor
            for the first n_times subsequent time steps,
            shape: (n_times,)

        Returns
        -------
        alpha: numpy.ndarray
            The delta WD result at the x locations,
            shape: (n_times,)

        """
        assert isinstance(algo, Sequential), (
            f"Wake deflection '{self.name}' requires Sequential algorithm, got '{type(algo).__name__}'"
        )

        n_times = len(x)

        def _get_data(var):
            data = algo.farm_results_downwind[var].to_numpy()[:n_times, downwind_index]
            data[-1] = fdata[var][0, downwind_index]
            return data

        gamma = _get_data(FV.YAWM)
        ct = _get_data(FV.CT)
        alpha = np.zeros_like(gamma)

        sel = (ct > 1e-8) & (np.abs(gamma) > 1e-8)
        if np.any(sel):
            D = _get_data(FV.D)[sel]
            gamma = np.deg2rad(gamma[sel])
            ct = ct[sel]

            alpha[sel] = np.rad2deg(
                -(np.cos(gamma) ** 2)
                * np.sin(gamma)
                * ct
                / 2
                / (1 + self.beta * x / D) ** 2
            )

        return alpha



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

from foxes import __version__

# -- Project information -----------------------------------------------------

project = "foxes"
copyright = "2025, Fraunhofer IWES"
author = "Fraunhofer IWES"

# The short X.Y version
version = __version__
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx_immaterial",
    "sphinx_immaterial.apidoc.python.apigen",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    # "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    # "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    # "sphinx.ext.inheritance_diagram",
    "sphinx.ext.doctest",
    "m2r2",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
}

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]
# autosummary_generate = False

# The suffix(es) of source filenames.
#
source_suffix = {".rst": "restructuredtext", ".md": "restructuredtext"}

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    # ipynb checkpoints
    "notebooks/.ipynb_checkpoints/*.ipynb",
    "build/*",
    # "_templates/*",
    # DEBUG
    # "examples.rst",
    # "notebooks/*",
    # "notebooks/layout_opt.ipynb",
    # "notebooks/timelines.ipynb",
    # "notebooks/heterogeneous.ipynb",
    # "notebooks/timeseries.ipynb",
    # "api.rst"
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None

# autosummary_generate = True
napolean_use_rtype = False

# -- Options for sphinxcontrib.email ------------------------------------------
# email_automode = True


# -- Options for nbsphinx -----------------------------------------------------

# Execute notebooks before conversion: 'always', 'never', 'auto' (default)
# We execute all notebooks, exclude the slow ones using 'exclude_patterns'
nbsphinx_execute = "always"

# Use this kernel instead of the one stored in the notebook metadata:
# nbsphinx_kernel_name = 'python3'

# List of arguments to be passed to the kernel that executes the notebooks:
# nbsphinx_execute_arguments = []

# If True, the build process is continued even if an exception occurs:
# nbsphinx_allow_errors = True


# Controls when a cell will time out (defaults to 30; use -1 for no timeout):
nbsphinx_timeout = 500

# Default Pygments lexer for syntax highlighting in code cells:
# nbsphinx_codecell_lexer = 'ipython3'

# Width of input/output prompts used in CSS:
# nbsphinx_prompt_width = '8ex'

# If window is narrower than this, input/output prompts are on separate lines:
# nbsphinx_responsive_width = '700px'

# This is processed by Jinja2 and inserted before each notebook
# Fix for issue with pyplot, cf
# https://github.com/readthedocs/sphinx_rtd_theme/issues/788#issuecomment-585785027
nbsphinx_prolog = r"""
.. raw:: html

    <script src='http://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.10/require.min.js'></script>
    <script>require=requirejs;</script>


"""

# This is processed by Jinja2 and inserted after each notebook
# nbsphinx_epilog = r"""
# """

# Input prompt for code cells. "%s" is replaced by the execution count.
nbsphinx_input_prompt = "In [%s]:"

# Output prompt for code cells. "%s" is replaced by the execution count.
nbsphinx_output_prompt = "Out[%s]:"

# Specify conversion functions for custom notebook formats:
# import jupytext
# nbsphinx_custom_formats = {
#    '.Rmd': lambda s: jupytext.reads(s, '.Rmd'),
# }

# Link or path to require.js, set to empty string to disable
# nbsphinx_requirejs_path = ''

# Options for loading require.js
# nbsphinx_requirejs_options = {'async': 'async'}
mathjax3_config = {
    "TeX": {"equationNumbers": {"autoNumber": "AMS", "useLabelIds": True}},
}

# Additional files needed for generating LaTeX/PDF output:
# latex_additional_files = ['references.bib']

# -- Options for autodoc ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_immaterial"

# html_theme = 'cloud'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # TOC options
    #'navigation_depth': 2,  # only show 2 levels on left sidebar
    # "collapse_navigation": False,  # don't allow sidebar to collapse,
    "site_url": "https://fraunhoferiwes.github.io/foxes.docs/index.html",
    "repo_url": "https://github.com/FraunhoferIWES/foxes",
    "icon": {"repo": "fontawesome/brands/github", "edit": "material/file-edit-outline"},
    "palette": {"primary": "teal"},
    "toc_title_is_page_title": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# custom.css is inside one of the html_static_path folders (e.g. _static)
# html_css_files = ["custom.css"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "foxesdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "foxes.tex", "foxes Documentation", "Fraunhofer IWES", "manual"),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "foxes", "foxes Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "foxes",
        "foxes Documentation",
        author,
        "foxes",
        "Farm Optimization and eXtended yield Evaluation Software",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- python_apigen configuration -------------------------------------------------

python_apigen_modules = {
    "foxes.variables": "_foxes/variables/",
    "foxes.constants": "_foxes/constants/",
    "foxes.algorithms": "_foxes/algorithms/",
    "foxes.algorithms.downwind": "_foxes/algorithms/downwind/",
    "foxes.algorithms.downwind.models": "_foxes/algorithms/downwind/models/",
    "foxes.algorithms.iterative": "_foxes/algorithms/iterative/",
    "foxes.algorithms.iterative.models": "_foxes/algorithms/iterative/models/",
    "foxes.algorithms.sequential": "_foxes/algorithms/sequential/",
    "foxes.algorithms.sequential.models": "_foxes/algorithms/sequential/models/",
    "foxes.config": "_foxes/config/",
    "foxes.core": "_foxes/core/",
    "foxes.data": "_foxes/data/",
    "foxes.engines": "_foxes/engines/",
    "foxes.input.farm_layout": "_foxes/input/farm_layout/",
    "foxes.input.states": "_foxes/input/states/",
    "foxes.input.states.create": "_foxes/input/states/create/",
    "foxes.input.yaml": "_foxes/input/yaml/",
    "foxes.input.yaml.windio": "_foxes/input/yaml/windio/",
    "foxes.output": "_foxes/output/",
    "foxes.output.flow_plots_2d": "_foxes/output/flow_plots_2d/",
    "foxes.output.seq_plugins": "_foxes/output/seq_plugins/",
    "foxes.models": "_foxes/models/",
    "foxes.models.farm_controllers": "_foxes/models/farm_controllers/",
    "foxes.models.farm_models": "_foxes/models/farm_models/",
    "foxes.models.partial_wakes": "_foxes/models/partial_wakes/",
    "foxes.models.point_models": "_foxes/models/point_models/",
    "foxes.models.rotor_models": "_foxes/models/rotor_models/",
    "foxes.models.turbine_models": "_foxes/models/turbine_models/",
    "foxes.models.turbine_types": "_foxes/models/turbine_types/",
    "foxes.models.vertical_profiles": "_foxes/models/vertical_profiles/",
    "foxes.models.axial_induction": "_foxes/models/axial_induction",
    "foxes.models.ground_models": "_foxes/models/ground_models",
    "foxes.models.wake_deflections": "_foxes/models/wake_deflections",
    "foxes.models.wake_frames": "_foxes/models/wake_frames/",
    "foxes.models.wake_models": "_foxes/models/wake_models/",
    "foxes.models.wake_models.induction": "_foxes/models/wake_models/induction/",
    "foxes.models.wake_models.wind": "_foxes/models/wake_models/wind/",
    "foxes.models.wake_models.ti": "_foxes/models/wake_models/ti/",
    "foxes.models.wake_superpositions": "_foxes/models/wake_superpositions/",
    "foxes.utils": "_foxes/utils/",
    "foxes.utils.abl": "_foxes/utils/abl",
    "foxes.utils.geom2d": "_foxes/utils/geom2d/",
    "foxes.utils.two_circles": "_foxes/utils/two_circles/",
    "foxes.utils.abl.neutral": "_foxes/utils/abl/neutral/",
    "foxes.utils.abl.stable": "_foxes/utils/abl/stable/",
    "foxes.utils.abl.unstable": "_foxes/utils/abl/unstable/",
    "foxes.utils.abl.sheared": "_foxes/utils/abl/sheared/",
}



import numpy as np
from scipy.interpolate import (
    LinearNDInterpolator,
    NearestNDInterpolator,
    RBFInterpolator,
)

from foxes.config import config
from foxes.utils import wd2uv, uv2wd, weibull_weights
import foxes.variables as FV
import foxes.constants as FC

from .dataset_states import DatasetStates


class PointCloudData(DatasetStates):
    """
    Inflow data with point cloud support.

    Attributes
    ----------
    states_coord: str
        The states coordinate name in the data
    point_coord: str
        The point coordinate name in the data
    x_ncvar: str
        The x variable name in the data
    y_ncvar: str
        The y variable name in the data
    h_ncvar: str, optional
        The height variable name in the data
    weight_ncvar: str, optional
        The name of the weights variable in the data
    interp_method: str
        The interpolation method, "linear", "nearest" or "radialBasisFunction"
    interp_fallback_nearest: bool
        If True, use nearest neighbor interpolation if the
        interpolation method fails.
    interp_pars: dict
        Additional arguments for the interpolation

    Examples
    --------
    Example of the NetCDF input files with point cloud data:

    >>>    Dimensions:  (point: 30, state: 100)
    >>>    Dimensions without coordinates: point, state
    >>>    Data variables:
    >>>        x        (point) float32 120B ...
    >>>        y        (point) float32 120B ...
    >>>        ws       (state, point) float32 12kB ...
    >>>        wd       (state, point) float32 12kB ...
    >>>        ti       (point) float32 120B ...
    >>>        rho      (state) float32 400B ...

    :group: input.states

    """

    def __init__(
        self,
        *args,
        states_coord="Time",
        point_coord="point",
        x_ncvar="x",
        y_ncvar="y",
        h_ncvar=None,
        weight_ncvar=None,
        interp_method="linear",
        interp_fallback_nearest=False,
        interp_pars={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        states_coord: str
            The states coordinate name in the data
        point_coord: str
            The point coordinate name in the data
        x_ncvar: str
            The x variable name in the data
        y_ncvar: str
            The y variable name in the data
        h_ncvar: str, optional
            The height variable name in the data
        weight_ncvar: str, optional
            The name of the weights variable in the data
        interp_method: str
            The interpolation method, "linear", "nearest" or "radialBasisFunction"
        interp_fallback_nearest: bool
            If True, use nearest neighbor interpolation if the
            interpolation method fails.
        interp_pars: dict
            Additional arguments for the interpolation
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(*args, **kwargs)

        self.states_coord = states_coord
        self.point_coord = point_coord
        self.x_ncvar = x_ncvar
        self.y_ncvar = y_ncvar
        self.h_ncvar = h_ncvar
        self.weight_ncvar = weight_ncvar
        self.interp_method = interp_method
        self.interp_pars = interp_pars
        self.interp_fallback_nearest = interp_fallback_nearest

        self.variables = [FV.X, FV.Y]
        self.variables += [v for v in self.ovars if v not in self.fixed_vars]
        self.var2ncvar[FV.X] = x_ncvar
        self.var2ncvar[FV.Y] = y_ncvar
        if weight_ncvar is not None:
            self.var2ncvar[FV.WEIGHT] = weight_ncvar
            self.variables.append(FV.WEIGHT)
        elif FV.WEIGHT in self.var2ncvar:
            raise KeyError(
                f"States '{self.name}': Cannot have '{FV.WEIGHT}' in var2ncvar, use weight_ncvar instead"
            )

        self._n_pt = None
        self._n_wd = None
        self._n_ws = None

        if FV.WS not in self.ovars:
            raise ValueError(
                f"States '{self.name}': Expecting output variable '{FV.WS}', got {self.ovars}"
            )
        if FV.WD not in self.ovars:
            raise ValueError(
                f"States '{self.name}': Expecting output variable '{FV.WD}', got {self.ovars}"
            )
        for v in [FV.WEIBULL_A, FV.WEIBULL_k, FV.WEIGHT]:
            if v in self.ovars:
                raise ValueError(
                    f"States '{self.name}': Cannot have '{v}' as output variable"
                )

        self._cmap = {
            FC.STATE: self.states_coord,
            FC.POINT: self.point_coord,
        }

    def __repr__(self):
        return f"{type(self).__name__}(n_pt={self._n_pt}, n_wd={self._n_wd}, n_ws={self._n_ws})"

    def _read_ds(self, ds, cmap, variables, verbosity=0):
        """
        Helper function for _get_data, extracts data from the original Dataset.

        Parameters
        ----------
        ds: xarray.Dataset
            The Dataset to read data from
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        coords: dict
            keys: Foxes variable names, values: 1D coordinate value arrays
        data: dict
            The extracted data, keys are variable names,
            values are tuples (dims, data_array)
            where dims is a tuple of dimension names and
            data_array is a numpy.ndarray with the data values

        """
        coords, data = super()._read_ds(ds, cmap, variables, verbosity)

        assert FV.X in data and FV.Y in data, (
            f"States '{self.name}': Expecting variables '{FV.X}' and '{FV.Y}' in data, found {list(data.keys())}"
        )
        assert data[FV.X][0] == (FC.POINT,), (
            f"States '{self.name}': Expecting variable '{FV.X}' to have dimensions '({FC.POINT},)', got {data[FV.X][0]}"
        )
        assert data[FV.Y][0] == (FC.POINT,), (
            f"States '{self.name}': Expecting variable '{FV.Y}' to have dimensions '({FC.POINT},)', got {data[FV.Y][0]}"
        )
        if FV.H in data:
            assert data[FV.H][0] == (FC.POINT,), (
                f"States '{self.name}': Expecting variable '{FV.H}' to have dimensions '({FC.POINT},)', got {data[FV.H][0]}"
            )

        points = [data.pop(FV.X)[1], data.pop(FV.Y)[1]]
        if FV.H in data:
            points.append(data.pop(FV.H)[1])
        coords[FC.POINT] = np.stack(points, axis=-1)

        return coords, data

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        return super().load_data(
            algo,
            cmap=self._cmap,
            variables=self.variables,
            bounds_extra_space=None,
            verbosity=verbosity,
        )

    def calculate(self, algo, mdata, fdata, tdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """

        # prepare
        self.ensure_output_vars(algo, tdata)
        n_states = tdata.n_states
        n_targets = tdata.n_targets
        n_tpoints = tdata.n_tpoints
        n_states = fdata.n_states
        n_pts = n_states * n_targets * n_tpoints
        coords = [self.states_coord, self.point_coord]

        # get data for calculation
        coords, data, weights = self.get_calc_data(mdata, self._cmap, self.variables)
        coords[FC.STATE] = np.arange(n_states, dtype=config.dtype_int)

        # interpolate data to points:
        out = {}
        for dims, (vrs, d) in data.items():
            # prepare
            n_vrs = len(vrs)
            qts = coords[FC.POINT]
            n_qts, n_dms = qts.shape
            idims = dims[:-1]

            if idims == (FC.STATE,):
                for i, v in enumerate(vrs):
                    if v in self.ovars:
                        out[v] = np.zeros(
                            (n_states, n_targets, n_tpoints), dtype=config.dtype_double
                        )
                        out[v][:] = d[:, None, None, i]
                continue

            elif idims == (FC.POINT,):
                # prepare grid data
                gts = qts
                n_gts = n_qts

                # prepare evaluation points
                pts = tdata[FC.TARGETS][..., :n_dms].reshape(n_pts, n_dms)

            elif idims == (FC.STATE, FC.POINT):
                # prepare grid data, add state index to last axis
                gts = np.zeros((n_qts, n_states, n_dms + 1), dtype=config.dtype_double)
                gts[..., :n_dms] = qts[:, None, :]
                gts[..., n_dms] = np.arange(n_states)[None, :]
                n_gts = n_qts * n_states
                gts = gts.reshape(n_gts, n_dms + 1)

                # reorder data, first to shape (n_qts, n_states, n_vars),
                # then to (n_gts, n_vrs)
                d = np.swapaxes(d, 0, 1)
                d = d.reshape(n_gts, n_vrs)

                # prepare evaluation points, add state index to last axis
                pts = np.zeros(
                    (n_states, tdata.n_targets, tdata.n_tpoints, n_dms + 1),
                    dtype=config.dtype_double,
                )
                pts[..., :n_dms] = tdata[FC.TARGETS][..., :n_dms]
                pts[..., n_dms] = np.arange(n_states)[:, None, None]
                pts = pts.reshape(n_pts, n_dms + 1)

            else:
                raise ValueError(
                    f"States '{self.name}': Unsupported dimensions {dims} for variables {vrs}"
                )

            # translate (WD, WS) to (U, V):
            if FV.WD in vrs or FV.WS in vrs:
                assert FV.WD in vrs and (FV.WS in vrs or FV.WS in self.fixed_vars), (
                    f"States '{self.name}': Missing '{FV.WD}' or '{FV.WS}' in data variables {vrs} for dimensions {dims}"
                )
                iwd = vrs.index(FV.WD)
                iws = vrs.index(FV.WS)
                ws = d[..., iws] if FV.WS in vrs else self.fixed_vars[FV.WS]
                d[..., [iwd, iws]] = wd2uv(d[..., iwd], ws, axis=-1)
                del ws

            # create interpolator
            if self.interp_method == "linear":
                interp = LinearNDInterpolator(gts, d, **self.interp_pars)
            elif self.interp_method == "nearest":
                interp = NearestNDInterpolator(gts, d, **self.interp_pars)
            elif self.interp_method == "radialBasisFunction":
                pars = {"neighbors": 10}
                pars.update(self.interp_pars)
                interp = RBFInterpolator(gts, d, **pars)
            else:
                raise NotImplementedError(
                    f"States '{self.name}': Interpolation method '{self.interp_method}' not implemented, choices are: 'linear', 'nearest', 'radialBasisFunction'"
                )

            # run interpolation
            ires = interp(pts)
            del interp

            # check for error:
            sel = np.any(np.isnan(ires), axis=-1)
            if np.any(sel):
                i = np.where(sel)[0]
                if self.interp_fallback_nearest:
                    interp = NearestNDInterpolator(gts, d)
                    pts = pts[i]
                    ires[i] = interp(pts)
                    del interp
                else:
                    p = pts[i[0], :n_dms]
                    qmin = np.min(qts[:, :n_dms], axis=0)
                    qmax = np.max(qts[:, :n_dms], axis=0)
                    raise ValueError(
                        f"States '{self.name}': Interpolation method '{self.interp_method}' failed for {np.sum(sel)} points, e.g. for point {p}, outside of bounds {qmin} - {qmax}"
                    )
            del pts, gts, d

            # translate (U, V) into (WD, WS):
            if FV.WD in vrs:
                uv = ires[..., [iwd, iws]]
                ires[..., iwd] = uv2wd(uv)
                ires[..., iws] = np.linalg.norm(uv, axis=-1)
                del uv

            # set output:
            for i, v in enumerate(vrs):
                out[v] = ires[..., i].reshape(n_states, n_targets, n_tpoints)
            del ires

        # set fixed variables:
        for v, d in self.fixed_vars.items():
            out[v] = np.full(
                (n_states, n_targets, n_tpoints), d, dtype=config.dtype_double
            )

        # add weights:
        if weights is not None:
            tdata[FV.WEIGHT] = weights[:, None, None]
        elif FV.WEIGHT in out:
            tdata[FV.WEIGHT] = out.pop(FV.WEIGHT)
        else:
            tdata[FV.WEIGHT] = np.full(
                (n_states, 1, 1), 1 / self._N, dtype=config.dtype_double
            )
        tdata.dims[FV.WEIGHT] = (FC.STATE, FC.TARGET, FC.TPOINT)

        return {v: out[v] for v in self.ovars}


class WeibullPointCloud(PointCloudData):
    """
    Weibull sectors at point cloud support, e.g., at turbine locations.

    Attributes
    ----------
    wd_coord: str
        The wind direction coordinate name
    ws_coord: str
        The wind speed coordinate name, if wind speed bin
        centres are in data, else None
    ws_bins: numpy.ndarray
        The wind speed bins, including
        lower and upper bounds, shape: (n_ws_bins+1,)

    Examples
    --------
    Example of the NetCDF input files with point cloud data:

    >>>    Dimensions:               (wind_turbine: 8, wind_direction: 2, wind_speed: 2)
    >>>    Coordinates:
    >>>    * wind_turbine          (wind_turbine) int64 64B 0 1 2 3 4 5 6 7
    >>>    * wind_direction        (wind_direction) int64 16B 0 30
    >>>    * wind_speed            (wind_speed) int64 16B 8 10
    >>>    Data variables:
    >>>        sector_probability    (wind_turbine, wind_direction) float64 128B ...
    >>>        weibull_a             (wind_turbine, wind_direction) float64 128B ...
    >>>        weibull_k             (wind_turbine, wind_direction) float64 128B ...
    >>>        turbulence_intensity  (wind_turbine, wind_direction, wind_speed) float64 256B ...
    >>>        x                     (wind_turbine) float64 64B ...
    >>>        y                     (wind_turbine) float64 64B ...
    >>>        height                (wind_turbine) float64 64B ...

    :group: input.states

    """

    def __init__(
        self,
        *args,
        wd_coord,
        ws_coord=None,
        ws_bins=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Positional arguments for the base class
        wd_coord: str
            The wind direction coordinate name
        ws_coord: str, optional
            The wind speed coordinate name, if wind speed bin
            centres are in data
        ws_bins: list of float, optional
            The wind speed bins, including
            lower and upper bounds
        kwargs: dict, optional
            Keyword arguments for the base class

        """
        super().__init__(
            *args,
            states_coord=wd_coord,
            time_format=None,
            load_mode="preload",
            **kwargs,
        )
        self.wd_coord = wd_coord
        self.ws_coord = ws_coord
        self.ws_bins = None if ws_bins is None else np.sort(np.asarray(ws_bins))

        assert ws_coord is not None or ws_bins is not None, (
            f"States '{self.name}': Expecting either ws_coord or ws_bins"
        )
        assert ws_coord is None or ws_bins is None, (
            f"States '{self.name}': Expecting either ws_coord or ws_bins, not both"
        )

        if FV.WD not in self.ovars:
            raise ValueError(
                f"States '{self.name}': Expecting output variable '{FV.WD}', got {self.ovars}"
            )
        for v in [FV.WEIBULL_A, FV.WEIBULL_k, FV.WEIGHT]:
            if v in self.ovars:
                raise ValueError(
                    f"States '{self.name}': Cannot have '{v}' as output variable"
                )
            if v not in self.variables:
                self.variables.append(v)

        for v in [FV.WS, FV.WD]:
            if v in self.variables:
                self.variables.remove(v)

        self._n_wd = None
        self._n_ws = None

    def __repr__(self):
        return f"{type(self).__name__}(n_wd={self._n_wd}, n_ws={self._n_ws})"

    def _read_ds(self, ds, cmap, variables, verbosity=0):
        """
        Helper function for _get_data, extracts data from the original Dataset.

        Parameters
        ----------
        ds: xarray.Dataset
            The Dataset to read data from
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        coords: dict
            keys: Foxes variable names, values: 1D coordinate value arrays
        data: dict
            The extracted data, keys are variable names,
            values are tuples (dims, data_array)
            where dims is a tuple of dimension names and
            data_array is a numpy.ndarray with the data values

        """
        # read data, using wd_coord as state coordinate
        hcmap = cmap.copy()
        if self.ws_coord is not None:
            hcmap = {FV.WS: self.ws_coord, **cmap}
        coords, data0 = super()._read_ds(ds, hcmap, variables, verbosity)
        wd = coords.pop(FC.STATE)
        wss = coords.pop(FV.WS, None)

        # replace state by wd coordinate
        data0 = {
            v: (tuple({FC.STATE: FV.WD}.get(c, c) for c in dims), d)
            for v, (dims, d) in data0.items()
        }

        # check weights
        if FV.WEIGHT not in data0:
            raise KeyError(
                f"States '{self.name}': Missing weights variable '{FV.WEIGHT}' in data, found {sorted(list(data0.keys()))}"
            )
        else:
            dims = data0[FV.WEIGHT][0]
            if FV.WD not in dims:
                raise KeyError(
                    f"States '{self.name}': Expecting weights variable '{FV.WEIGHT}' to contain dimension '{FV.WD}', got {dims}"
                )
            if FV.WS in dims:
                raise KeyError(
                    f"States '{self.name}': Expecting weights variable '{FV.WEIGHT}' to not contain dimension '{FV.WS}', got {dims}"
                )

        # construct wind speed bins and bin deltas
        assert FV.WS not in data0, (
            f"States '{self.name}': Cannot have '{FV.WS}' in data, found variables {list(data0.keys())}"
        )
        if self.ws_bins is not None:
            wsb = self.ws_bins
            wss = 0.5 * (wsb[:-1] + wsb[1:])
        elif wss is not None:
            wsb = np.zeros((len(wss) + 1,), dtype=config.dtype_double)
            wsb[1:-1] = 0.5 * (wss[1:] + wss[:-1])
            wsb[0] = wss[0] - 0.5 * wsb[1]
            wsb[-1] = wss[-1] + 0.5 * wsb[-2]
            self.ws_bins = wsb
        else:
            raise ValueError(
                f"States '{self.name}': Expecting ws_bins argument, or '{self.ws_coord}' among data coordinates, got {list(ds.coords.keys())}"
            )
        wsd = wsb[1:] - wsb[:-1]
        n_ws = len(wss)
        n_wd = len(wd)
        del wsb

        # calculate Weibull weights
        dms = [FV.WS, FV.WD]
        shp = [n_ws, n_wd]
        for v in [FV.WEIBULL_A, FV.WEIBULL_k]:
            if FC.POINT in data0[v][0]:
                dms.append(FC.POINT)
                shp.append(data0[v][1].shape[data0[v][0].index(FC.POINT)])
                break
        dms = tuple(dms)
        shp = tuple(shp)
        if data0[FV.WEIGHT][0] == dms:
            w = data0.pop(FV.WEIGHT)[1]
        else:
            s_w = tuple([np.s_[:] if c in data0[FV.WEIGHT][0] else None for c in dms])
            w = np.zeros(shp, dtype=config.dtype_double)
            w[:] = data0.pop(FV.WEIGHT)[1][s_w]
        s_ws = tuple([np.s_[:], None] + [None] * (len(dms) - 2))
        s_A = tuple([np.s_[:] if c in data0[FV.WEIBULL_A][0] else None for c in dms])
        s_k = tuple([np.s_[:] if c in data0[FV.WEIBULL_A][0] else None for c in dms])
        data0[FV.WEIGHT] = (
            dms,
            w
            * weibull_weights(
                ws=wss[s_ws],
                ws_deltas=wsd[s_ws],
                A=data0.pop(FV.WEIBULL_A)[1][s_A],
                k=data0.pop(FV.WEIBULL_k)[1][s_k],
            ),
        )
        del w, s_ws, s_A, s_k

        # translate binned data to states
        self._N = n_ws * n_wd
        self._inds = None
        data = {
            FV.WS: np.zeros((n_ws, n_wd), dtype=config.dtype_double),
            FV.WD: np.zeros((n_ws, n_wd), dtype=config.dtype_double),
        }
        data[FV.WS][:] = wss[:, None]
        data[FV.WD][:] = wd[None, :]
        data[FV.WS] = ((FC.STATE,), data[FV.WS].reshape(self._N))
        data[FV.WD] = ((FC.STATE,), data[FV.WD].reshape(self._N))
        for v in list(data0.keys()):
            dims, d = data0.pop(v)
            if len(dims) >= 2 and dims[:2] == (FV.WS, FV.WD):
                dms = tuple([FC.STATE] + list(dims[2:]))
                shp = [self._N] + list(d.shape[2:])
                data[v] = (dms, d.reshape(shp))
            elif dims[0] == FV.WD:
                dms = tuple([FC.STATE] + list(dims[1:]))
                shp = [n_ws] + list(d.shape)
                data[v] = np.zeros(shp, dtype=config.dtype_double)
                data[v][:] = d[None, ...]
                data[v] = (dms, data[v].reshape([self._N] + shp[2:]))
            elif dims[0] == FV.WS:
                dms = tuple([FC.STATE] + list(dims[1:]))
                shp = [n_ws, n_wd] + list(d.shape[2:])
                data[v] = np.zeros(shp, dtype=config.dtype_double)
                data[v][:] = d[:, None, ...]
                data[v] = (dms, data[v].reshape([self._N] + shp[2:]))
            else:
                data[v] = (dims, d)

        return coords, data



import numpy as np
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    # define arguments and options:
    parser = argparse.ArgumentParser()
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.08)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-dx", help="The turbine spacing in x", type=float, default=800.0
    )
    parser.add_argument(
        "-dy", help="The turbine spacing in y", type=float, default=400.0
    )
    parser.add_argument(
        "-nx", help="The number of turbines in x direction", type=int, default=6
    )
    parser.add_argument(
        "-ny", help="The number of turbines in y direction", type=int, default=10
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["RHB", "Bastankhah2014_linear_lim_k004"],
        nargs="+",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default="rotor_wd")
    parser.add_argument("-v", "--var", help="The plot variable", default=FV.WS)
    parser.add_argument("-e", "--engine", help="The engine", default=None)
    parser.add_argument(
        "-nit",
        "--not_iterative",
        help="Don't use the iterative algorithm",
        action="store_true",
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    # create model book
    mbook = foxes.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    # create states
    states = foxes.input.states.SingleStateStates(
        ws=args.ws, wd=args.wd, ti=args.ti, rho=args.rho
    )

    # create wind farm
    print("\nCreating wind farm")
    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_grid(
        farm=farm,
        xy_base=[0.0, 0.0],
        step_vectors=[[args.dx, 0.0], [0.0, args.dy]],
        steps=[args.nx, args.ny],
        turbine_models=["DTU10MW"],
        verbosity=0,
    )

    # create algorithm
    Algo = (
        foxes.algorithms.Downwind if args.not_iterative else foxes.algorithms.Iterative
    )
    algo = Algo(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame=args.frame,
        partial_wakes=args.pwakes,
        mbook=mbook,
        engine=args.engine,
        verbosity=1,
    )

    # calculate farm results
    farm_results = algo.calc_farm()
    print("\nResults data:\n", farm_results)

    # add capacity and efficiency to farm results
    o = foxes.output.FarmResultsEval(farm_results)
    o.add_capacity(algo)
    o.add_capacity(algo, ambient=True)
    o.add_efficiency()

    # state-turbine results
    farm_df = farm_results.to_dataframe()
    print("\nFarm results data:\n")
    print(
        farm_df[
            [
                FV.X,
                FV.WD,
                FV.AMB_REWS,
                FV.REWS,
                FV.AMB_TI,
                FV.TI,
                FV.AMB_P,
                FV.P,
                FV.CT,
                FV.EFF,
            ]
        ]
    )
    print()

    # results by turbine
    turbine_results = o.reduce_states(
        {
            FV.AMB_P: "weights",
            FV.P: "weights",
            FV.AMB_CAP: "weights",
            FV.CAP: "weights",
        }
    )
    turbine_results[FV.AMB_YLD] = o.calc_turbine_yield(
        algo=algo, annual=True, ambient=True
    )
    turbine_results[FV.YLD] = o.calc_turbine_yield(algo=algo, annual=True)
    turbine_results[FV.EFF] = turbine_results[FV.P] / turbine_results[FV.AMB_P]
    print("\nResults by turbine:\n")
    print(turbine_results)

    # power results
    P0 = o.calc_mean_farm_power(ambient=True)
    P = o.calc_mean_farm_power()
    print(f"\nFarm power        : {P / 1000:.1f} MW")
    print(f"Farm ambient power: {P0 / 1000:.1f} MW")
    print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
    print(f"Annual farm yield : {turbine_results[FV.YLD].sum():.2f} GWh.")

    # horizontal flow plot
    if not args.nofig:
        o = foxes.output.FlowPlots2D(algo, farm_results)
        xmin = -2000
        xmax = (args.nx - 1) * args.dx + 2000
        g = o.gen_states_fig_xy(
            args.var,
            figsize=(7, 5),
            resolution=20,
            xmin=xmin,
            xmax=xmax,
            yspace=1000.0,
            levels=40,
            rotor_color="black",
        )
        fig = next(g)
        plt.show()
        plt.close(fig)

        # center line plot:
        H = mbook.turbine_types[ttype.name].H
        n_points = 10000
        points = np.zeros((1, n_points, 3))
        points[:, :, 0] = np.linspace(xmin, xmax, n_points)[None, :]
        points[:, :, 2] = H
        point_results = algo.calc_points(farm_results, points)
        fig, ax = plt.subplots(figsize=(10, 5))
        ax.plot(points[0, :, 0], point_results[FV.WS][0, :])
        ax.set_xlabel("x [m]")
        ax.set_ylabel("Wind speed [m/s]")
        plt.show()
        plt.close(fig)

        # front line plot:
        points = np.zeros((1, n_points, 3))
        points[:, :, 0] = -200
        points[:, :, 1] = np.linspace(-500.0, args.ny * args.dy + 500, n_points)[
            None, :
        ]
        points[:, :, 2] = H
        point_results = algo.calc_points(farm_results, points)
        fig, ax = plt.subplots(figsize=(10, 5))
        ax.plot(point_results[FV.WS][0, :], points[0, :, 1])
        ax.set_ylabel("y [m]")
        ax.set_xlabel("Wind speed [m/s]")
        plt.show()
        plt.close(fig)
